/*

######################################################################
######################################################################
##                                                                  ##
##                                                                  ##
##                            statnet                               ##
##                                                                  ##
##            a feedforward interpolation neural network            ##
##                                                                  ##
##                              by                                  ##
##                                                                  ##
##                       Coryn Bailer-Jones                         ##
##                                                                  ##
##                           07/12/98                               ##
##                                                                  ##
##             email: calj@mpia-hd.mpg.de                           ##
##               www: http://www.mpia-hd.mpg.de/stars/calj/         ##
##                                                                  ##
##                                                                  ##
##        see the README file for disclaimer and warranty           ##
##        see the statnet_manual file for operational details       ##
##                                                                  ##
##     This file is copyright 1998,1999 by C.A.L. Bailer-Jones      ##
##                                                                  ##
##                                                                  ##
######################################################################
######################################################################


FILE:		statnet.c
DESCRIPTION:    ANSI C statnet program
AUTHOR:		Coryn Bailer-Jones
LAST MOD DATE:	24.06.99


######################################################################
######################################################################
##                                                                  ##
##                           statnet.c                              ##
##                                                                  ##
######################################################################
######################################################################


  Flags: (X = not yet implemented) (O = obsolete)

  1-49 inc. are user defined flags
  51+       are program internal flags 

  ********* User defined flags **********
  
  flag[1]  (train?) 0 = no, 1 = yes
  flag[2]  (apply?) 0 = no, 1 = yes
  flag[3]  (optimization method) 1 = gradient descent
                                 2 = macopt (forces flag[4]=3)
  flag[4]  (update method) 1 = svu (single vector update)
                           2 = rsu (random sample update)
			   3 = batch
X flag[5]  (error function) 1 = sum of squares			   
  flag[7]  (Use beta (noise) parameters?) 0 = no
                                          1 = yes and set to defaults
					  2 = yes and read in from file
  flag[8]  (Use alpha (weight decay) parameters?) 0 = no
                                                  1 = yes and set to defaults
						  2 = yes and read in from file
  flag[9]  (nature of weight initialisation) 1 = uniform distribution
                                             2 = gaussian distribution
  flag[10] (read in weights?) 0 = no  (=> initialise weights)
                              1 = yes (set if there's a line in spec file
			               giving input weight file name)
  flag[11] (output weights file specified?) 0 = no
                   (-> give warning and use default weights file) 
                                            1 = yes
                   (-> set if there's a line in spec file giving
	               output weight file name)
  flag[12] (data scaling method) 0 = none
                                 1 = zero mean, unit standard deviation
				 2 = maxmin (not yet implemented)
				 3 = only use Hlam (i.e. partial scaling by
				     number of inputs)
  flag[13] (perform maccheckgrad?) 0 = no, 1 = yes
  flag[14] (verbosity level)       0 = nothing, except intro title and errors
                                       (warnings are suppressed)
				       (will force flag[13]=0)
				   1 = minimum
				   2 = normal 
				   3 = routine calls & diagnostic information
				   4 = everything
				       (lots of run specific screen dumps
				        currently written in at this level)
  flag[16] (write errors at each iteration to a file?) 0 = no, 1 = yes
  flag[17] (write results for each network in committee?) 0 = no, 1 = yes
  flag[18] (number of patterns to exclude) 0 = don't exclude any
                                          >0 is then number to exclude
	    
  ********* Internal program flags **********
				 
  flag[52] (has network just been trained?)
                0 = no  
		1 = yes
  flag[53] (number of hidden layers) 
                must be 1 or 2 (no default)
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "statnet.h"



/* =================== Start of Main Program ================== */

void main(int numinvar, char *invar[])
{
  /* declarations */

  int flag[Nflags];
  int *xranges[2]; /* see below in allocation for explanation */ 
  int *yranges[2];
  
  int Nnets; /* number of networks in the committee */
  int Xsize; /* number of input nodes (does not include bias) */
  int Hsize; /* number of nodes in first hidden layer (does not include bias) */
  int Vsize; /* number of nodes in second hidden layer (does not include bias) */
             /* note that Vsize=0 in the specfile means use only one hiddden 
		layer of size Hsize, but Vsize does not then keep it's
		"correct value". See "NOTE ON 1 AND 2 HIDDEN LAYER MODELS" below.
	     */
  int Ysize; /* number of output nodes */
  int Xtot;  /* total number of input  variables in data files */
  int Ytot;  /* total number of output variables in data files */
  int Nits;  /* number of training iterations (if doing graddescent) */
  long ranseed; /* or do we want to use a pointer instead? */
  
  int    macitmax;    /* maximum number of iterations to use in macopt */
  double macconvtol;  /* convergence tolerance in macopt */
  double macchecktol; /* gradient checking tolerance in maccheckgrad */
  
  double eta;   /* step size in gradient descent */
  double wtrng; /* range of initial weights */ 
  
  double ***wtXH; /* weights for XH layer : 
		     wtXH[n][l][m] - network, input, first hidden */
  double ***wtHV; /* weights for HV layer : 
		     wtHV[n][m][j] - network, first hidden, second hidden */
  double ***wtVY; /* weights for VY layer :
		     wtVY[n][j][k] - network, second hidden, output */

  /* NOTE ON 1 AND 2 HIDDEN LAYER MODELS
     
     If we are only using a network with one hidden layer (i.e. flag[53]=1)
     then wtVY is the hidden-output weights matrix.
     wtHV is still allocated memory and passed to some subroutines, but 
     is not used as such. This may seem wasteful, but I want to use
     some dedicated code for the one hidden layer case, and don't want to
     write a separate program. flag[53] is set in specread(), where
     we also set Vsize=Hsize for the 1 hidden layer case. However,
     when we write to an output file (e.g. the weights file) we write
     Vsize=0, so we know that it's a one hidden layer network. Thus
     we must be careful, i.e. check flag[53] when we compare the value
     of Vsize within the code with Vsize read in from a weights (or
     other) file.
  */

  int takelog[MAXOUTS+1]; 
  /* takelog[f]=k, for k>=0 means that we should work with the log
     (base 10) of target variable f, where f is the number (starting
     at 1) of the output as it occurs in the data file, and k is its
     position in the selected data range. These values are assigned in
     dataread().  takelog[f]=-1 is an initial flag to say that we will
     take the log, but don't know yet what the value of k is: these
     are assigned in specread().  takelog[f]=-2 means we don't take
     the log at all. This is the default value. I use a fix-sized array
     to avoid grief!
     */
  
  double *yscale[2], *xscale[2];
  /* scale parameters for v and x respectively
     if flag[12]=1: scale[0] contains mean (double)
		    scale[1] contains standard deviations (double) */
  double Hlam,Vlam; /* lambda parameters in tanh transfer function
		       It is passed to/from statnettrain and statnetapply as a
		       pointer as statnetapply will use weights and scale
		       factors from statnettrain if we do the training */

  char *exclude[MAXEXCLUDE];    /* ids of patterns to exclude */

  char trnfname[FNAMESIZE];      /* name of training data file */
  char appfname[FNAMESIZE];      /* name of application data file */
  char sfname[FNAMESIZE];        /* specfile name */
  char inwtfname[FNAMESIZE];     /* input weight file name */
  char otwtfname[FNAMESIZE];     /* output weight file name */
  char errfname[FNAMESIZE];      /* error file name */

  int i,p;

  /* Anfang */

  printf("\n*** statnet version %s last modified %s ***\n\n", VERSION, LASTMODDATE);
  
  if(numinvar < 2)
    neterr("main(): must specify name of specfile on command line");
  if(strcmp(invar[1],"-v")==0)
    exit(0);

  printf("Program started: ");
  fflush(stdout);
  system("date"); /* UNIX specific */
  fflush(stdout);
  
  /* initialise fixed arrays: int takeflag[MAXOUTS+1]={-2} didn't work */

  for(i=0;i<Nflags;i++) flag[i]=0;  
  for(i=1;i<=MAXOUTS;i++) takelog[i]=-2;

  /* allocate memory  - I wanted to use static 2D arrays, but kept getting
     unexplained incompatible pointer types when passing to functions. */
  xranges[0]=Ivector(MAXRANGES+1);
  xranges[1]=Ivector(MAXRANGES+1);
  yranges[0]=Ivector(MAXRANGES+1);
  yranges[1]=Ivector(MAXRANGES+1);

  /* for each r>0, xranges[0][r] to xranges[1][r] stores the inclusive
     range of the input vector to be read. xranges[0][0] stores the 
     number of such ranges. xranges[1][0] is not used. Thus if you 
     want only the first 10 inputs, you would set: 
     xranges[0][0]=1 
     xranges[0][1]=1     NOT 0 
     xranges[1][1]=10    NOT 9
     same applies for yranges
     */

  /* obtain and check specifications (set defaults and read specfile) */

  strcpy(sfname,invar[1]);
  specread(sfname, flag, &Nnets, &Xsize, &Hsize, &Vsize, &Ysize, &Xtot, &Ytot,
	   xranges, yranges, &Nits, &eta, trnfname, appfname,
	   inwtfname, otwtfname, &ranseed, &macconvtol,
	   &macitmax, &macchecktol, &wtrng, errfname, takelog, exclude);
    
  /* allocate memory */

  wtXH = Dmatrix3(Nnets, Xsize+1, Hsize);
  wtHV = Dmatrix3(Nnets, Hsize+1, Vsize);
  wtVY = Dmatrix3(Nnets, Vsize+1, Ysize);

  yscale[0] = Dvector(Ysize);
  yscale[1] = Dvector(Ysize);
  xscale[0] = Dvector(Xsize);
  xscale[1] = Dvector(Xsize);
		     
  /* the business */

  flag[52]=0;
  if(flag[1]==1) {
    statnettrain(flag, Nnets, Xsize, Hsize, Vsize, Ysize, 
		 Xtot, Ytot, xranges, yranges,
		 wtXH, wtHV, wtVY, yscale, xscale, &Hlam, &Vlam, trnfname,
		 Nits, eta, sfname, inwtfname, otwtfname, ranseed,
		 macconvtol, macitmax, macchecktol, wtrng, errfname,
		 takelog, exclude);
    flag[52]=1;
  }
  if(flag[2]==1)
    statnetapply(flag, Nnets, Xsize, Hsize, Vsize, Ysize, 
		 Xtot, Ytot, xranges, yranges,
		 wtXH, wtHV, wtVY, yscale, xscale, &Hlam, &Vlam, appfname,
		 sfname, inwtfname, otwtfname, ranseed, wtrng, 
		 takelog, exclude); 

  /* deallocations */

  for(p=0;p<flag[18];++p)
    free_Cvector(exclude[p],0);  /* allocated in specread */

  free_Ivector(xranges[0], MAXRANGES+1);
  free_Ivector(xranges[1], MAXRANGES+1);
  free_Ivector(yranges[0], MAXRANGES+1);
  free_Ivector(yranges[1], MAXRANGES+1);

  free_Dmatrix3(wtXH, Nnets, Xsize+1, Hsize);
  free_Dmatrix3(wtHV, Nnets, Hsize+1, Vsize);
  free_Dmatrix3(wtVY, Nnets, Vsize+1, Ysize);

  free_Dvector(xscale[0], Xsize);
  free_Dvector(xscale[1], Xsize);
  free_Dvector(yscale[0], Ysize);
  free_Dvector(yscale[1], Ysize);

  if(flag[14]>=1) printf("\n*** statnet completed ***\n\n");
  exit(0);
  
}

/* ==================== End of Main Program =================== */



/* ==================== Principal Control Routines =================== */  

void statnettrain(int *flag, int Nnets, 
		  int Xsize, int Hsize, int Vsize, int Ysize,
		  int Xtot, int Ytot, int **xranges, int **yranges,
		  double ***wtXH, double ***wtHV, double ***wtVY,
		  double **yscale, double **xscale, double *Hlam, double *Vlam,
		  char *trnfname, int Nits, double eta,
		  char *sfname, char *inwtfname, char *otwtfname, long ranseed,
		  double macconvtol, int macitmax, double macchecktol,
		  double wtrng, char *errfname, int *takelog, char *exclude[])
{
  /* train statnet */


  int n,f,k;
  int Nexcluded=0; /* only required by deallocation of tar and x */
  int Npats;     /* number of training patterns: read in by dataread()
		    from the header of the data file,
		    (total number of patterns in file) but then modified in 
		    dataread() to be number of patterns used if
		    we exclude patterns (flag[18]). Number excluded
		    (returned by dataread() for a given file is Nexcluded.
		    */

  double **x;    /* input data, x[p][l] (pattern,node) */
  targets **tar; /* defined in statnet.h: [p][k] (pattern,node) */
                 /* x and tar are allocated in dataread() */

  /* Note that memory is not assigned for either x or tar in this function,
     so that their base pointers are not assigned to the requried piece
     of memory. Thus you cannot call dataread() just by passing the base
     pointers (i.e. passing x and tar): C passes by value, so you'd just be
     passing the value of the base pointer which is then changed in dataread()
     when you allocate memory. Rather you want to pass p, a pointer to the
     base pointers, so that dataread() can change the contents of p (i.e.
     what p points to). You achieve this in the usual way, i.e. by passing 
     &x and &tar.
  */


  double *alpha; /* Bayes' weight decay parameters */
  double *beta;  /* noise levels */
  

  if(flag[14]>=2) printf("statnettrain(): Training network\n");

  /* read in data */

  dataread(Xsize, Ysize, Xtot, Ytot, xranges, yranges, 
	   &Npats, trnfname, &x, &tar, takelog, flag, exclude, &Nexcluded);
  if(flag[14]>=2) { 
    for(f=1;f<=Ytot;++f) {
      if(takelog[f]>-1)
	printf("Output variable %d will be logged as requested\n", f);
    }
  }

  /* allocate memory */

  alpha = Dvector(NALPHA);
  beta  = Dvector(Ysize);
  
  /* initialise committee */

  cominit(Nnets, Xsize, Hsize, Vsize, Ysize, wtXH, wtHV, wtVY, yscale, xscale,
	  Hlam, Vlam, flag, inwtfname, sfname, ranseed, alpha, beta, wtrng);

  if(flag[10]==0 && flag[12]!=0)
    scalecalc(Xsize, Hsize, Ysize, Npats, x, tar, xscale, yscale, Hlam, Vlam, flag);

  if(flag[12]==1 || flag[12]==2)
    datascale(Xsize, Ysize, Npats, x, tar, xscale, yscale, flag);
  
  /* train */

  /* loop over each network in committee. Subroutines evaluate weights
     for each network separately (hence the passing of wtXH[n] rather than
     wtXH */
     
  for(n=0;n<Nnets;++n) {
    if(flag[14]>=2) 
      printf("statnettrain(): Training network %d of %d\n",n+1,Nnets);
    if(flag[3]==1) { /* gradient descent */
      graddescent(Xsize, Hsize, Vsize, Ysize, wtXH[n], wtHV[n], wtVY[n], x, tar, 
		  Npats, alpha, beta, *Hlam, *Vlam, flag, Nits, eta, errfname);
    }
    else if(flag[3]==2) { /* macopt */
      callmacopt(Xsize, Hsize, Vsize, Ysize, wtXH[n], wtHV[n], wtVY[n], x,
		 tar, Npats, alpha, beta, *Hlam, *Vlam, flag,
		 errfname, macconvtol, macitmax, macchecktol);
    }
    else neterr("statnettrain(): flag[3] option does not exist");
  }  
  
  /* write out weights */

  if(flag[11]==1)
    writeweights(Nnets, Xsize, Hsize, Vsize, Ysize, wtXH, wtHV, wtVY, 
		 yscale, xscale, *Hlam, *Vlam, otwtfname, flag);
  
  /* deallocate memory */

  free_Dmatrix(x,Npats+Nexcluded,Xsize+1);
  free_tmatrix(tar,Npats+Nexcluded,Ysize);

  free_Dvector(alpha,NALPHA);
  free_Dvector(beta,Ysize);

}



void statnetapply(int *flag, int Nnets, 
		  int Xsize, int Hsize, int Vsize, int Ysize,
		  int Xtot, int Ytot, int **xranges, int **yranges, 
		  double ***wtXH, double ***wtHV, double ***wtVY,
		  double **yscale, double **xscale, double *Hlam, double *Vlam,
		  char *appfname, char *sfname,
		  char *inwtfname, char *otwtfname, long ranseed, 
		  double wtrng, int *takelog, char *exclude[])
{
  /* Apply network */

  int p,n,k,f;
  int Nexcluded=0;
  int Npats;     /* number of patterns (input-output pairs)
		    - see note under declaration of Npats in statnettrain() */
  
  double **x;    /* input data, x[p][l] (pattern,node) */
  targets **tar; /* defined in statnet.h: [p][k] (pattern,node) */
                 /* x and tar are allocated in dataread() */

  double *h;   
  double *v;
  double **y;       /* network outputs, y[n][k] (net,node) */   
  double *comclass; /* committee's outputs, comclass[k] (node) */
  double *thistar;
  double *alpha;  /* alpha and beta not needed in this routine but */
  double *beta;   /* required for compatibility with cominit */ 
  double *interr; /* internal error for output vector of a single pattern */
                  /* defined by eqn.1 in thesis p1 */
  double *avinterr;     /* average of interr across all patterns */
  double *rmsexterr;    /* RMS external error across all patterns */
  double *avabsexterr;  /* average absolute external error across all patterns */
  int    *numdef;       /* for each output node, number of patterns with
			   a defined target */

  FILE *tpotf;
  char tpotname[FNAMESIZE];
  char wtfname[FNAMESIZE]; /* weights file name to write to tpot file,
			    i.e. the weights used in this net application */
  char message[MESSAGESIZE];


  if(flag[14]>=2) printf("statnetapply(): Entering network application phase\n");
  
  /* read in data */

  dataread(Xsize, Ysize, Xtot, Ytot, xranges, yranges,
	   &Npats, appfname, &x, &tar, takelog, flag, exclude, &Nexcluded);
  if(flag[14]>=2) { 
    for(f=1;f<=Ytot;++f) {
      if(takelog[f]>-1)
	printf("Output variable %d will be logged as requested\n", f);
    }
  }


  /* allocate memory */

  if(flag[14]>=4) printf("POSITION 01 reached\n");

  h =           Dvector(Hsize+1); 
  v =           Dvector(Vsize+1);
  y =           Dmatrix(Nnets,Ysize); 
  comclass =    Dvector(Ysize); 
  thistar =     Dvector(Ysize); 
  interr  =     Dvector(Ysize); 
  avinterr =    Dvector(Ysize); 
  rmsexterr =   Dvector(Ysize); 
  avabsexterr = Dvector(Ysize); 
  numdef =      Ivector(Ysize); 
  alpha =       Dvector(NALPHA);
  beta  =       Dvector(Ysize);  

  if(flag[14]>=4) printf("POSITION 05 reached\n");

  /* initialise arrays and committee error measures */

  h[Hsize]=HIDBIAS; 
  v[Vsize]=HIDBIAS;
  for(k=0;k<Ysize;++k) {
    avinterr[k]=0.0;
    rmsexterr[k]=0.0;
    avabsexterr[k]=0.0;
    numdef[k]  =0.0;
  } 

  /* do scaling */

  /* need to scale tar when not training.  Note that if we are
     applying a network with random weights, then tar is never scaled
     so scale values of 0.0 will be written to the output file. We
     could scale them, but it would be meaningless: no dataset has
     been used to obtain the weights so what dataset do we use to
     obtain the scale? This would change if we ever do teacher forcing
     in the application. */
  
  if(flag[52]==0) { /* => we don't have weights or scale factors,
		       so we'll either read weights in or use random ones */
    cominit(Nnets, Xsize, Hsize, Vsize, Ysize, wtXH, wtHV, wtVY, yscale, xscale, 
	    Hlam, Vlam, flag, inwtfname, sfname, ranseed, alpha, beta, wtrng);  
    if(flag[10]==0 && flag[12]!=0) /* => are using random weights and
				         want to scale data */
      scalecalc(Xsize, Hsize, Ysize, Npats, x, tar, xscale, yscale, Hlam, Vlam, flag);
    if(flag[10]==1) /* => have read in weights */
      strcpy(wtfname,inwtfname); 
    else /* => are using random weights */
      strcpy(wtfname,otwtfname); 
  }
  else /* => are using weights and scale factors found from training */
    strcpy(wtfname,otwtfname); 
  
  if(flag[12]==1 || flag[12]==2)
      datascale(Xsize, Ysize, Npats, x, tar, xscale, yscale, flag);
  
  /* apply each network in turn to each pattern */

  if(flag[14]>=2)
    printf("statnetapply(): Applying Committee\n");
  
  evnewname(tpotname,appfname); /* evaluate tpotname */
  tpotf=fopen(tpotname,"w");
  if(!tpotf) {
    sprintf(message, "statnetapply(): output file %s could not be opened", tpotname);
    neterr(message);
  }
  if(flag[14]>=2)
    printf("Input data file name: %s\nClassifications output file name: %s\n", 
	   appfname, tpotname);

  fprintf(tpotf, "# statnet classifications output file\n");
  fprintf(tpotf, "# ###################################\n");
  fprintf(tpotf, "# input   file = %s\n", appfname);
  fprintf(tpotf, "# weights file = %s\n", wtfname);
  fprintf(tpotf, "# N (nets), Y (outputs)\n");
  fprintf(tpotf, "  %d         %d\n", Nnets, Ysize);
  fprintf(tpotf,"#");
  for(k=0;k<Ysize;++k) {
    fprintf(tpotf, "Com     True    Diff   IntErr  ");
    if(flag[17]==1) {
      for(n=0;n<Nnets;++n) 
	fprintf(tpotf, "Net#%d  ", n+1);
    }
  }
  fprintf(tpotf, "\n");

  for(p=0;p<Npats;++p) { /* loop over patterns */
    if(flag[14]>=3) printf("classifying pattern no. %2d\n", p+1);

    for(k=0;k<Ysize;++k) { /* initialise */
      comclass[k]=0.0;
      interr[k]=0.0;
      thistar[k] = tar[p][k].val;
    }    

    if(flag[12]==1 || flag[12]==2)
      unscale(Ysize, thistar, yscale, flag);

    for(n=0;n<Nnets;++n) { /* loop over networks to evaluate y */
      if(flag[53]==1)
	forpass(Xsize, Hsize, Ysize, x[p], h, y[n], wtXH[n], wtVY[n], 
		*Hlam, flag);
      else if(flag[53]==2)
	forpass2(Xsize, Hsize, Vsize, Ysize, x[p], h, v, y[n], 
		 wtXH[n], wtHV[n], wtVY[n], *Hlam, *Vlam, flag);
      else
	neterr("statnetapply(): flag[53] not recognised");
      if(flag[12]==1 || flag[12]==2)
	unscale(Ysize, y[n], yscale, flag);
      for(k=0;k<Ysize;++k)
	comclass[k]+=y[n][k];
    } /* end of networks */
    for(k=0;k<Ysize;++k)
      comclass[k]/=(float)Nnets;

    for(k=0;k<Ysize;++k) { /* contribute conclass to sum squared external error */
      if(tar[p][k].def==1) {
	rmsexterr[k]+=(comclass[k]-thistar[k])*(comclass[k]-thistar[k]);
	avabsexterr[k]+=fabs(comclass[k]-thistar[k]);
	numdef[k]++;
      }
    }

    if(Nnets>1) { /* evaluate internal error */
      for(k=0;k<Ysize;++k) { 
	if(tar[p][k].def==1) {
	  for(n=0;n<Nnets;++n)
	    interr[k]+=(y[n][k]-comclass[k])*(y[n][k]-comclass[k]);
	}
	interr[k]/=(float)(Nnets-1);
	interr[k]=sqrt(interr[k]);
	if(tar[p][k].def==1) 
	  avinterr[k]+=interr[k];
      }
    }
    
    for(k=0;k<Ysize;++k) { /* write results for each output in turn */
      fprintf(tpotf, "% 6.4f % 6.4f % 6.4f %6.4f ", comclass[k], thistar[k], comclass[k]-thistar[k], interr[k]);
      if(flag[17]==1) {
	for(n=0;n<Nnets;++n)
	  fprintf(tpotf, "%6.4f ", y[n][k]);
      }
    } /* end of output nodes */
    fprintf(tpotf, "\n");

  } /* end of patterns */

  /* report statistics */

  if(flag[14]>=1) {
    printf("\nOutput   No_def_outputs   RMS_ext_error   Av_abs_ext_error   Av_int_error\n");
    for(k=0;k<Ysize;++k) { 
      if(numdef[k]>2) 
	rmsexterr[k]/=(float)(numdef[k]-1);
      rmsexterr[k]=sqrt(rmsexterr[k]);
      avabsexterr[k]/=(float)numdef[k];
      avinterr[k]/=(float)numdef[k];
      printf("  %2d          %4d           %6.4f           %6.4f            %6.4f\n", k+1, numdef[k], rmsexterr[k], avabsexterr[k], avinterr[k]);
    }
  }

  /* write out weights if we've been using random weights */

  if(flag[11]==1 && flag[52]==0 && flag[10]==0)
    writeweights(Nnets, Xsize, Hsize, Vsize, Ysize, wtXH, wtHV, wtVY,
		 yscale, xscale, *Hlam, *Vlam, otwtfname, flag);

  /* deallocate memory - x and tar allocated in dataread() */

  free_Dmatrix(x,Npats+Nexcluded,Xsize+1); 
  free_tmatrix(tar,Npats+Nexcluded,Ysize);
  free_Dvector(h,Hsize+1);
  free_Dvector(v,Vsize+1);
  free_Dmatrix(y,Nnets,Ysize);
  free_Dvector(comclass,Ysize);
  free_Dvector(thistar,Ysize);
  free_Dvector(interr,Ysize);
  free_Dvector(avinterr,Ysize);
  free_Dvector(rmsexterr,Ysize);
  free_Dvector(avabsexterr,Ysize);
  free_Ivector(numdef,Ysize);
  free_Dvector(alpha,NALPHA);
  free_Dvector(beta,Ysize);
 
}



/* ==================== Forward Pass Routines =================== */

void forpass(int Xsize, int Hsize, int Ysize, double *x, double *h, double *y, 
	     double **wtXH, double **wtHY, double Hlam, int *flag)
{
  /* forward pass through the network for one hidden layer.
     i.e. given x and the weights (inc. Hlam), evaluate h and y.

     Note that hidden-output weights are called HY here, whereas they are
     VY in the "mother" subroutines.
     */
  
  int k,m;

  double *p; /* h = f(p) */
  double *q; /* y = g(q) */

  p = Dvector(Hsize);
  q = Dvector(Ysize);


  if(flag[14]>=4) {
    for(k=0;k<Xsize+1;++k)
      printf("forpass(): x[%2d] = %f\n", k, x[k]);
  }

  vectmat(wtXH, x, Xsize+1, Hsize, p);
  for(m=0;m<Hsize;++m) {
    h[m] = Tf(p[m], Hlam);
    if(flag[14]>=4)
      printf("forpass(): p[%d] = %f,  h[%d] = %f\n", m,p[m],m,h[m]);
  }

  vectmat(wtHY, h, Hsize+1, Ysize, q);
  for(k=0;k<Ysize;++k) /* NB for linear outputs Tg(q[n]) = q[n] */
    y[k] = Tg(q[k]);
    
  free_Dvector(p, Hsize);
  free_Dvector(q, Ysize);
  
}



void forpass2(int Xsize, int Hsize, int Vsize, int Ysize, 
	      double *x, double *h, double *v, double *y, 
	      double **wtXH, double **wtHV, double **wtVY,
	      double Hlam, double Vlam, int *flag)
{
  /* forward pass through the network for two hidden layers.
     i.e. given x and the weights (inc. Hlam), evaluate h, v and y.

     */
  
  int m,j,k;

  double *p; /* h = f(p) */
  double *q; /* v = f(q) */
  double *r; /* y = g(r) */

  p = Dvector(Hsize);
  q = Dvector(Vsize);
  r = Dvector(Ysize);

  vectmat(wtXH, x, Xsize+1, Hsize, p);
  for(m=0;m<Hsize;++m)
    h[m] = Tf(p[m], Hlam);

  vectmat(wtHV, h, Hsize+1, Vsize, q);
  for(j=0;j<Vsize;++j)
    v[j] = Tf(q[j], Vlam);

  vectmat(wtVY, v, Vsize+1, Ysize, r);
  for(k=0;k<Ysize;++k) /* NB for linear outputs Tg(r[n]) = r[n] */
    y[k] = Tg(r[k]);
    
  free_Dvector(p, Hsize);
  free_Dvector(q, Vsize);
  free_Dvector(r, Ysize);
  
}



/* ==================== Gradient Descent Routines =================== */

void graddescent(int Xsize, int Hsize, int Vsize, int Ysize,
		 double **wtXH, double **wtHV, double **wtVY, double **x,
		 targets **tar, int Npats,
		 double *alpha, double *beta, double Hlam, double Vlam, 
		 int *flag, int Nits, double eta, char *errfname)
{
  /* optimise weights by gradient descent for finite and fixed
     number of iterations 
     */

  FILE *errf;
  char message[MESSAGESIZE];
  float lerr;   /* likelihood error */
  float werr;   /* weight decay error */
  float toterr; /* lerr+werr */
  float temp;

  double **edXH; /* error derivatives wrt weights */
  double **edHV;
  double **edVY;
  double **cumedXH; /* cumulative of ed */
  double **cumedHV;
  double **cumedVY;

  double *h;    /* first hidden layer */
  double *v;    /* second hidden layer */
  double *y;    /* output layer */
  
  int n,p,l,m,j,k;

  if(flag[14]>=3) printf("graddescent(): Training network with gradient descent\n");

  /* allocate memory */
  
  edXH    = Dmatrix(Xsize+1, Hsize); /* dE/dw_ij */
  edHV    = Dmatrix(Hsize+1, Vsize);
  edVY    = Dmatrix(Vsize+1, Ysize);
  cumedXH = Dmatrix(Xsize+1, Hsize); /* Sum dE/dw_ij */
  cumedHV = Dmatrix(Hsize+1, Vsize);
  cumedVY = Dmatrix(Vsize+1, Ysize);
  h       = Dvector(Hsize+1); 
  v       = Dvector(Vsize+1); 
  y       = Dvector(Ysize); 

  /* initialise */

  h[Hsize]=HIDBIAS; /* initialises hidden layer bias node */
  v[Vsize]=HIDBIAS; /* initialises hidden layer bias node */

  for(m=0;m<Hsize;++m) {
    for(l=0;l<Xsize+1;++l)
      cumedXH[l][m]=0.0;
  }
  if(flag[53]==2) {
    for(m=0;m<Hsize+1;++m) {
      for(j=0;j<Vsize;++j)
	cumedHV[m][j]=0.0;
    }  
  }
  for(j=0;j<Vsize+1;++j) {
    for(k=0;k<Ysize;++k)
      cumedVY[j][k]=0.0;
  }  
  
  if(flag[16]==1) { /* we want to write errors to a file during training */
    errf=fopen(errfname,"w");
    if(!errf) {
      sprintf(message, "callmacopt(): error file %s could not be opened", errfname);
      neterr(message);
    }
  }
  
  for(n=1;n<=Nits;++n) { /* iterations */
    lerr=0.0;
    werr=0.0;

    for(p=0;p<Npats;++p) { /* patterns */
      if(flag[14]>=4) printf("graddescent(): p=%2d\n", p);
      if(flag[53]==1) {
	forpass(Xsize, Hsize, Ysize, x[p], h, y, wtXH, wtVY, Hlam, flag);
	ederiv(Xsize, Hsize, Ysize, x[p], h, y, wtXH, wtVY, edXH, edVY, 
	       tar[p], alpha, beta, Hlam, flag); 
      }
      else if(flag[53]==2) {
	forpass2(Xsize, Hsize, Vsize, Ysize, x[p], h, v, y, 
		 wtXH, wtHV, wtVY, Hlam, Vlam, flag);
	ederiv2(Xsize, Hsize, Vsize, Ysize, x[p], h, v, y, 
		wtXH, wtHV, wtVY, edXH, edHV, edVY, 
		tar[p], alpha, beta, Hlam, Vlam, flag); 
      }
      else
	neterr("graddescent(): flag[53] not recognised");
      if(flag[4]==1) /* svu - passes edXH, edHV and edVY to updatewt() */
	updatewt(Xsize, Hsize, Vsize, Ysize, wtXH, wtHV, wtVY, 
		 edXH, edHV, edVY, eta, flag);
      else if(flag[4]==3)
	cumederivs(Xsize, Hsize, Vsize, Ysize, edXH, edHV, edVY, 
		   cumedXH, cumedHV, cumedVY, flag);
      else
	neterr("graddescent(): weight update method (flag[4]) does not exist");
      
      for(k=0;k<Ysize;++k) { /* only contribute to error if target defined */
	if(tar[p][k].def) { 
	  temp = errfn(y[k],tar[p][k].val);
	  lerr += 0.5*beta[k]*temp*temp;
	}
      }
      if(flag[8]>0) /* weight decay contribution to error */
	werr += wderr(Xsize, Hsize, Vsize, Ysize, wtXH, wtHV, wtVY, alpha, flag); 
      toterr=werr+lerr;
      
    } /* end of pattern loop */
    
    if(flag[14]>=4) printf("graddescent(): Got to end of pattern loop\n");
    if(flag[4]==3) /* batch - passes cumedXH, cumedHV and cumedVY to updatewt() */
      updatewt(Xsize, Hsize, Vsize, Ysize, wtXH, wtHV, wtVY, 
	       cumedXH, cumedHV, cumedVY, eta, flag);
    if(flag[14]>=4) printf("graddescent(): Have updated weights\n");
    if(flag[16]==1)
      fprintf(errf, "%5d  %.5e %5.1f  %.4e %5.1f   %.5e\n", n, lerr, 
	      100*lerr/toterr, werr, 100*werr/toterr, toterr);
    if(flag[14]>=2) 
      printf("%2d of %d its: L=%.5e (%3.1f%%) W=%.5e (%3.1f%%) Tot=%.5e\n", 
	     n, Nits, lerr, 100*lerr/toterr, werr, 100*werr/toterr, toterr);

  } /* end of iteration loop */



  /* deallocate memory */

  free_Dmatrix(edXH, Xsize+1, Hsize);
  free_Dmatrix(edHV, Hsize+1, Vsize);
  free_Dmatrix(edVY, Vsize+1, Ysize);
  free_Dmatrix(cumedXH, Xsize+1, Hsize);
  free_Dmatrix(cumedHV, Hsize+1, Vsize);
  free_Dmatrix(cumedVY, Vsize+1, Ysize);
  free_Dvector(h, Hsize+1);
  free_Dvector(v, Vsize+1);
  free_Dvector(y, Ysize);

  if(flag[16]==1)
    fclose(errf);

  if(flag[14]>=4) printf("graddescent(): Got to end\n");  

}



void updatewt(int Xsize, int Hsize, int Vsize, int Ysize,
	      double **wtXH, double **wtHV, double **wtVY, 
	      double **cumedXH, double **cumedHV, double **cumedVY, 
	      double eta, int *flag)
{
  /* do a gradient descent update of the weights and set cumed to zero.
     handles 1 and 2 hidden layers.
     
     if we're doing batch update then we pass in the edXH etc. parameters
     rather than the cumedXH etc. parameters.

     Note that we only bother to re-initialise cumulated variables 
     if we've actually used them before, i.e. if we're not doing SVU.
     (If we are doing SVU, then the ed variables are always initialised
     at the beginning of ederiv() anyway.)
     */

  int l,m,j,k;

  for(m=0;m<Hsize;++m) { /* not including hidden layer bias */
    for(l=0;l<Xsize+1;++l) {
      wtXH[l][m] += -eta*cumedXH[l][m];
      if(flag[4]!=1)
	cumedXH[l][m]=0.0;
    }
  }

  if(flag[53]==2) {
    for(m=0;m<Hsize+1;++m) { 
      for(j=0;j<Vsize;++j) {
	wtHV[m][j] += -eta*cumedHV[m][j];
	if(flag[4]!=1)
	  cumedHV[m][j]=0.0;
      }
    }  
  }
  
  for(j=0;j<Vsize+1;++j) { 
    for(k=0;k<Ysize;++k) {
      wtVY[j][k] += -eta*cumedVY[j][k];
      if(flag[4]!=1)
	cumedVY[j][k]=0.0;
    }
  }  

}



/* ==================== Macopt-relevant Routines =================== */

void callmacopt(int Xsize, int Hsize, int Vsize, int Ysize,
		double **wtXH, double **wtHV, double **wtVY,
		double **x, targets **tar, 
		int Npats, double *alpha, double *beta, double Hlam,
		double Vlam, int *flag, char *errfname,
		double macconvtol, int macitmax, double macchecktol)
{

  /* macoptII is called, which does the entire minimization itself.

     macopt_defaults(&mca);
     - This sets default values for the macopt parameters held in
       the structure mca
       
     macoptII(wtvec, wtvec_size, statmacint, (void *)(&dga), &mca);
     - wtvec is the initial vector from which minimization commences;
       statmacint is the name of the function which macoptII must call
       in order to evaluate the gradient of the error function;
       dga is a structure of parameters which statmacint needs;
       statmacint is generally called many times by macoptII.
       Upon returning from macoptII, wtvec holds the optimized weights
       
     maccheckgrad(wtvec, wtvec_size, (double)macchecktol,
                  statmacfn, (void *)(&dga), statmacint, (void *)(&dga), 0); 
     - maccheckgrad tests to see whether the gradient of the error
       function is correct by comparing it with an approximation to
       the gradient obtained from a first difference method using the
       function itself. maccheckgrad must be passed the names of two
       functions, the first which evaluates the function (statmacfn),
       and the second which evaluates its derivatives (statmacint).
       statmacfn is basically the same as statmacint but with the
       derivative stuff removed and the with the correct arguments to
       be compatiable with macopt.  macchecktol is the tolerance to
       which first differences are calculated. This is a bit
       repetitive, but as maccheckgrad is not an integral part of the
       code (it's only for debugging) it's not worth worrying
       about. The final zero just ensures that checking is done for
       all weights.
   */
  
  statmacint_args dga; /* defined in statnet.h */
  macopt_args mca;     /* defined in macopt.h */
  
  double *wtvec;  /* concatenated weight vector for passing to macopt
		     As we do pointer arithmetic on this, type must be
		     consistent with usage elsewhere (i.e. in statmacint) */
  int l,m,j,k;    /* counters */
  int offset;
  int wtvec_size;
  char message[MESSAGESIZE];
  
  if(flag[14]>=3) printf("callmacopt(): Using macopt to train network\n");
  
  /* allocate memory */

  wtvec_size = (Xsize+1)*Hsize + (Vsize+1)*Ysize;
  if(flag[53]==2)
    wtvec_size+=(Hsize+1)*Vsize;
  wtvec = Dvector(wtvec_size);


  /* Convert weight matrices into single vector for macopt.  Elements
     of wtvec are in order wtXH, wtHV, wtVY. If we're only using one
     hidden layer, then all of wtHV weights are omitted.

     This wastes a little time as it is repeated every time macoptII
     calls statmacint. It also increases memory usage as wtvec is copied
     to new wt matrices in statmacint. The alternative, however, is to
     use a vector for the weights throughout, which is prone to
     producing bugs given the intrinsic matrix nature of the weights.
     Memory isn't a problem as the weight matrices are insignificant
     compared to the error derivative arrays. Time also isn't much of
     an issue as it's just copying the matrices once per call of
     statmacint. */

  offset = 0;
  for(l=0;l<Xsize+1;++l) {
    for(m=0;m<Hsize;++m)
      wtvec[offset + Hsize*l + m] = wtXH[l][m];
  }
  offset += (Xsize+1)*Hsize;
  if(flag[53]==2) {
    for(m=0;m<Hsize+1;++m) {
      for(j=0;j<Vsize;++j)
	wtvec[offset + Vsize*m + j] = wtHV[m][j];
    }
    offset += (Hsize+1)*Vsize;
  }     
  for(j=0;j<Vsize+1;++j) {
    for(k=0;k<Ysize;++k)
      wtvec[offset + Ysize*j + k] = wtVY[j][k];
  }
  
  /* assign values of dga */
  
  dga.Xsize = Xsize;
  dga.Hsize = Hsize;
  dga.Vsize = Vsize;
  dga.Ysize = Ysize;
  dga.alpha = alpha;
  dga.beta = beta;
  dga.x = x;
  dga.tar = tar;
  dga.Hlam = Hlam;
  dga.Vlam = Vlam;
  dga.Npats = Npats;
  dga.flag = flag;
  dga.wtvec_size = wtvec_size;
  dga.gtyp = &(mca.gtyp);
  mca.its = 0; /* macopt does not initialise this before its first call of
		  statmacint(). I initialise it to prevent a line of garbage 
		  being written to the .err file */
  dga.its = &(mca.its);
  dga.previts = 0;
  if(flag[16]==1) { /* we want to write errors to a file during training */
    dga.errf=fopen(errfname,"w");
    if(!dga.errf) {
      sprintf(message, "callmacopt(): error file %s could not be opened", errfname);
      neterr(message);
    }
  }
  
  /* macopt stuff - see top of subroutine for details */
  
  wtvec--; /* macopt uses single offset vectors */
  macopt_defaults(&mca); /* set macopt parameters to defaults... */
  mca.verbose = flag[14]-1; 
  mca.tol = macconvtol; /* ...except for those from specfile */
  mca.itmax = macitmax;
  mca.rich = 1;
  mca.end_if_small_step = 0;  /* i.e. macopt will only convergence
				 once the gradient is small enough.
				 It will not converge simply if the
				 step is small */
  if(flag[13]==1) { /* check gradient prior to training */
    if(flag[14]>=3) printf("callmacopt(): Calling maccheckgrad\n");
    maccheckgrad(wtvec, wtvec_size, macchecktol,
		 statmacfn, (void *)(&dga), statmacint, (void *)(&dga), 0);
  }
  macoptII(wtvec, wtvec_size, statmacint, (void *)(&dga), &mca);
  if(flag[13]==1) { /* check gradient after training */
    if(flag[14]>=3) printf("callmacopt(): Calling maccheckgrad\n");
    maccheckgrad(wtvec, wtvec_size, macchecktol,
		 statmacfn, (void *)(&dga), statmacint, (void *)(&dga), 0);
  }
  wtvec++;

  /* rewrite wtvec back into matrices */

  offset = 0;
  for(l=0;l<Xsize+1;++l) {
    for(m=0;m<Hsize;++m) {
      wtXH[l][m] = wtvec[offset + Hsize*l + m];
      if(flag[14]>=4) printf("wtXH[%d][%d] = %8.5f ", l, m, wtXH[l][m]);      
    }
    if(flag[14]>=4) printf("\n");
  }
  offset += (Xsize+1)*Hsize;
  if(flag[53]==2) {
    for(m=0;m<Hsize+1;++m) {
      for(j=0;j<Vsize;++j) {
	wtHV[m][j] = wtvec[offset + Vsize*m + j];
	if(flag[14]>=4) printf("wtHV[%d][%d] = %8.5f ", m, j, wtHV[m][j]);
      }
      if(flag[14]>=4) printf("\n");
    }
    offset += (Hsize+1)*Vsize;
  }     
  for(j=0;j<Vsize+1;++j) {
    for(k=0;k<Ysize;++k) {
      wtVY[j][k] = wtvec[offset + Ysize*j + k];
      if(flag[14]>=4) printf("wtVY[%d][%d] = %8.5f ", j, k, wtVY[j][k]);
    }
    if(flag[14]>=4) printf("\n");
  }

    
  /* deallocate memory */

  free_Dvector(wtvec, wtvec_size);
  
  /* close files */
  
  if(flag[16]==1)
    fclose(dga.errf);
  
}



void statmacint(double *wtvec, double *wt_grad, void *pass_args)
{
  /* This is the interface between statnet and macopt: it is the
     gradient evaluating function which macopt calls.  wt_grad is
     declared and defined (i.e. memory allocated) in macopt.c.  The
     wtvec passed in may be the initial vectors (declared and defined
     in callmacopt) or later vectors (declared and defined in macopt).
     
     wtvec (vector) is the weights (passed into this function)
     wt_grad (vector) is the gradient of weights (evaluated by this function) 
     pass_args (structure - see below) carries the arguments for this function

     i.e. only wt_grad is changed by this function.
     
     Note that macopt defines (and/or uses) wtvec and wt_grad as unit
     offset (i.e they start at 1 rather than 0 like in most C code) so
     the pointers which it passes don't actually point to the
     beginning of the vector but one place earlier. To get around this
     we increment the pointers at the beginning of this function and
     decrement them before leaving. Thus the pointer wtvec and wt_grad
     should only be used locally. As we do pointer arithmetic, types
     for these two pointers have to be consistent with those in macopt
     and callmacopt. The alternative is to put "+1" wherever wtvec
     and wt_grad are used.

     pass_args is a pointer to the dga structure defined and declared
     in callmacopt. This pointer (and the data therein) is not
     affected by macopt.

     Note that macopt calls this function more than once per
     iteration.  At the beginning of iteration n, macopt prints the
     gradient from the end of the *previous* iteration. Therefore, in
     this function (below) when an iteration change is detected at the
     beginning of the function, the error evaluated at the end of the
     previous function call (stored as dga->lerr and dga->werr) are
     reported and associated with the current iteration number. This
     way the error reported by this function and the gradient reported
     by macopt correspond to the same iteration and the same point on
     the minimization surface. Note that the final gradient is not
     reported by macopt, and hence the final error is not reported
     either. (It macopt does not hit the iteration limit, this
     function does not get called so does not even write the
     penultimate error.)

     Note that the first time macopt calls this function, *(dga->its)
     has not been defined, so is garbage.
     
     Batch update is currently hard-wired into this algorithm.
     Other update methods are not easily compatible with macopt.
     */

  statmacint_args *dga = (statmacint_args *) pass_args; 
  /* This is necessary to be compatible with macopt. Format taken from
     MacKay's test_function.c */
       
  double **wtXH;    /* weights */
  double **wtHV;
  double **wtVY;
  double **edXH;    /* error derivatives wrt weights */
  double **edHV;
  double **edVY;
  double **cumedXH; /* cumulative of ed */
  double **cumedHV;
  double **cumedVY;

  double *h;
  double *v;
  double *y;

  double toterr;  /* the total error (lerr+werr) */
  double grad;    /* grad is the magnitude of the total gradient.
		     gg from macopt is sum of squares of gradient.
		     gtyp is from macopt, where gtyp = sqrt (gg/wtvec_size)
		     grad = sqrt (gg) = gtyp * sqrt (wtvec_size) */
  double temp;
  
  int p,l,m,j,k;
  int offset;

  if(dga->flag[14]>=3)
    printf("statmacint(): Performing forward & backward passes\n");

  if(dga->flag[14]>=4) printf("statmacint(): dga->its=%p, *(dga->its)=%d, dga->previts=%d\n", dga->its, *(dga->its), dga->previts);

  if(*(dga->its)!=dga->previts) { /* i.e. if we're on the next iteration */
    toterr=dga->lerr+dga->werr;
    grad = *(dga->gtyp)*sqrt((double)dga->wtvec_size);
    if(dga->flag[16]==1) /* file opened and closed in callmacopt() */
      fprintf(dga->errf, "%5d  %.5e %5.1f  %.4e %5.1f   %.5e   %.5e\n", *(dga->its), dga->lerr, 100*dga->lerr/toterr, dga->werr, 100*dga->werr/toterr, toterr, grad);
    if(dga->flag[14]>=3)
      printf("%d: Errors: L = %.5e (%3.1f%%)  W = %.5e (%3.1f%%)  Tot = %.5e\n", *(dga->its), dga->lerr, 100*dga->lerr/toterr, dga->werr, 100*dga->werr/toterr, toterr);
  }
  dga->lerr = 0.0;
  dga->werr = 0.0;

  /* allocate memory */

  if(dga->flag[14]>=4) printf("statmacint(): Xsize=%d, Hsize=%d, Vsize=%d, Ysize=%d,\n", dga->Xsize, dga->Hsize, dga->Vsize, dga->Ysize);

  wtXH    = Dmatrix(dga->Xsize+1, dga->Hsize); /* weights */
  wtHV    = Dmatrix(dga->Hsize+1, dga->Vsize);
  wtVY    = Dmatrix(dga->Vsize+1, dga->Ysize);
  edXH    = Dmatrix(dga->Xsize+1, dga->Hsize); /* dE/dw_ij */
  edHV    = Dmatrix(dga->Hsize+1, dga->Vsize);
  edVY    = Dmatrix(dga->Vsize+1, dga->Ysize);
  cumedXH = Dmatrix(dga->Xsize+1, dga->Hsize); /* Sum dE/dw_ij */
  cumedHV = Dmatrix(dga->Hsize+1, dga->Vsize);
  cumedVY = Dmatrix(dga->Vsize+1, dga->Ysize);

  h = Dvector(dga->Hsize+1);
  v = Dvector(dga->Vsize+1);
  y = Dvector(dga->Ysize);

  /* increment pointers to accommodate the unit offset in macopt */

  wtvec++;
  wt_grad++;
  
  /* initialise */
  
  offset = 0;
  for(l=0;l<dga->Xsize+1;++l) {
    for(m=0;m<dga->Hsize;++m)
      wtXH[l][m] = wtvec[offset + dga->Hsize*l + m];
  }
  offset += (dga->Xsize+1)*dga->Hsize;
  if(dga->flag[53]==2) {
    for(m=0;m<dga->Hsize+1;++m) {
      for(j=0;j<dga->Vsize;++j)
	wtHV[m][j] = wtvec[offset + dga->Vsize*m + j];
    }
    offset += (dga->Hsize+1)*dga->Vsize;
  }     
  for(j=0;j<dga->Vsize+1;++j) {
    for(k=0;k<dga->Ysize;++k)
      wtVY[j][k] = wtvec[offset + dga->Ysize*j + k];
  }

  h[dga->Hsize]=HIDBIAS; /* initialises hidden layer bias node */
  v[dga->Vsize]=HIDBIAS; /* initialises hidden layer bias node */

  for(m=0;m<dga->Hsize;++m) {
    for(l=0;l<dga->Xsize+1;++l)
      cumedXH[l][m]=0.0;
  }
  if(dga->flag[53]==2) {
    for(m=0;m<dga->Hsize+1;++m) {
      for(j=0;j<dga->Vsize;++j)
	cumedHV[m][j]=0.0;
    }  
  }
  for(j=0;j<dga->Vsize+1;++j) {
    for(k=0;k<dga->Ysize;++k)
      cumedVY[j][k]=0.0;
  }  
  
  /* Using batch update, evaluate error derivatives */
  
  for(p=0;p<dga->Npats;++p) { /* patterns */
    if(dga->flag[53]==1) {
      forpass(dga->Xsize, dga->Hsize, dga->Ysize, dga->x[p], h, y, 
	      wtXH, wtVY, dga->Hlam, dga->flag);
      ederiv(dga->Xsize, dga->Hsize, dga->Ysize,
	     dga->x[p], h, y, wtXH, wtVY, edXH, edVY, dga->tar[p],
	     dga->alpha, dga->beta, dga->Hlam, dga->flag); 
    }
    else if(dga->flag[53]==2) {
      forpass2(dga->Xsize, dga->Hsize, dga->Vsize, dga->Ysize, 
	       dga->x[p], h, v, y, wtXH, wtHV, wtVY, 
	       dga->Hlam, dga->Vlam, dga->flag);
      ederiv2(dga->Xsize, dga->Hsize, dga->Vsize, dga->Ysize,
	      dga->x[p], h, v, y, wtXH, wtHV, wtVY, edXH, edHV, edVY, 
	      dga->tar[p], dga->alpha, dga->beta, dga->Hlam, dga->Vlam, dga->flag); 
    }
    else
      neterr("statmacint(): flag[53] not recognised");
    if(dga->flag[14]>=4) {
      printf("p=%2d: wtXH = %10.7f, wtVY = %10.7f\n", p, wtXH[0][0], wtVY[0][0]);
      printf("dga->Xsize=%d dga->Hsize=%d dga->Ysize=%d\n", dga->Xsize, dga->Hsize, dga->Ysize);
    }
    cumederivs(dga->Xsize, dga->Hsize, dga->Vsize, dga->Ysize, 
	       edXH, edHV, edVY, cumedXH, cumedHV, cumedVY, dga->flag);
    /* diagnostics */
    if(dga->flag[14]>=4) {
      printf("statmacint(): edXH = %10.7f, edHV = %10.7f\n", edXH[0][0], edHV[0][0]);
    }
    for(k=0;k<dga->Ysize;++k) { 
      /* only contribute to error if target defined */
      if(dga->tar[p][k].def) { 
	temp = errfn(y[k],dga->tar[p][k].val);
	dga->lerr += 0.5*dga->beta[k]*temp*temp;
      }
    }
    if(dga->flag[8]>0) /* weight decay contribution to error */
      dga->werr += wderr(dga->Xsize, dga->Hsize, dga->Vsize, dga->Ysize,
			 wtXH, wtHV, wtVY, dga->alpha, dga->flag); 
  } /* end of pattern loop */


  toterr=dga->lerr+dga->werr;
  /*
  if(dga->flag[14]>=3)
    printf("Errors: L = %.5e (%3.1f%%)  W = %.5e (%3.1f%%)  Tot = %.5e\n", dga->lerr, 100*dga->lerr/toterr, dga->werr, 100*dga->werr/toterr, toterr);
  */
  dga->previts = *(dga->its);
  
  /* copy cumed into wt_grad */

  offset = 0;
  for(l=0;l<dga->Xsize+1;++l) {
    for(m=0;m<dga->Hsize;++m)
      wt_grad[offset + dga->Hsize*l + m] = cumedXH[l][m];
  }
  offset += (dga->Xsize+1)*dga->Hsize;
  if(dga->flag[53]==2) {
    for(m=0;m<dga->Hsize+1;++m) {
      for(j=0;j<dga->Vsize;++j)
	wt_grad[offset + dga->Vsize*m + j] = cumedHV[m][j];
    }
    offset += (dga->Hsize+1)*dga->Vsize;
  }     
  for(j=0;j<dga->Vsize+1;++j) {
    for(k=0;k<dga->Ysize;++k)
      wt_grad[offset + dga->Ysize*j + k] = cumedVY[j][k];
  }

  /* deallocate memory */
  
  free_Dmatrix(wtXH, dga->Xsize+1, dga->Hsize);
  free_Dmatrix(wtHV, dga->Hsize+1, dga->Vsize);
  free_Dmatrix(wtVY, dga->Vsize+1, dga->Ysize);
  free_Dmatrix(edXH, dga->Xsize+1, dga->Hsize);
  free_Dmatrix(edHV, dga->Hsize+1, dga->Vsize);
  free_Dmatrix(edVY, dga->Vsize+1, dga->Ysize);
  free_Dmatrix(cumedXH, dga->Xsize+1, dga->Hsize);
  free_Dmatrix(cumedHV, dga->Hsize+1, dga->Vsize);
  free_Dmatrix(cumedVY, dga->Vsize+1, dga->Ysize);

  free_Dvector(h, dga->Hsize+1);
  free_Dvector(v, dga->Vsize+1);
  free_Dvector(y, dga->Ysize);

  /* decrement pointers to accommodate the unit offset in macopt */

  wtvec--;
  wt_grad--;

}


double statmacfn(double *wtvec, void *pass_args)
{
  /* Evaluates the error function for a given weight vector (wtvec)
     and returns it as the function value. Neither wtvec nor
     pass_args are changed by this function.
     
     This is only used by maccheckgrad.  It is very similar to
     statmacint, but with derivative code removed.  Error is evaluated
     from scratch, i.e. using all the input x values etc. in
     pass_args. This is inefficient, but it doesn't matter as it's
     only a diagnostic.  Ultimately you could probably integrate this
     with statnetapply.

     Note that if we use maccheckgrad before using macoptII, 
     *(dga->its) is undefined, so will give garbage.

     batch update (flag[4]=3) hardwired into this function.
     */

  statmacint_args *dga = (statmacint_args *) pass_args; 
  /* This is necessary to be compatible with macopt. Format taken from
     MacKay's test_function.c */

  double lerr = 0.0; /* the likelihhod contribution to the error */
  double werr = 0.0; /* the weight decay contribution to the error */
  double toterr;     /* the total error (lerr+werr) */
  double temp;
  
  double **wtXH;  /* weights */
  double **wtHV;
  double **wtVY;

  double *h;      /* first hidden layer */
  double *v;      /* second hidden layer */
  double *y;      /* output layer */
  
  int p;
  int l,m,j,k;
  int offset;


  if(dga->flag[14]>=3)
    printf("statmacfn(): Performing forward pass\n");

  /* allocate memory */
  
  wtXH = Dmatrix(dga->Xsize+1, dga->Hsize);
  wtHV = Dmatrix(dga->Hsize+1, dga->Vsize);
  wtVY = Dmatrix(dga->Vsize+1, dga->Ysize);

  h    = Dvector(dga->Hsize+1);
  v    = Dvector(dga->Vsize+1);
  y    = Dvector(dga->Ysize);

  /* increment pointer to accommodate the unit offset in macopt */

  wtvec++;
  
  /* initialise */

  offset = 0;
  for(l=0;l<dga->Xsize+1;++l) {
    for(m=0;m<dga->Hsize;++m)
      wtXH[l][m] = wtvec[offset + dga->Hsize*l + m];
  }
  offset += (dga->Xsize+1)*dga->Hsize;
  if(dga->flag[53]==2) {
    for(m=0;m<dga->Hsize+1;++m) {
      for(j=0;j<dga->Vsize;++j)
	wtHV[m][j] = wtvec[offset + dga->Vsize*m + j];
    }
    offset += (dga->Hsize+1)*dga->Vsize;
  }     
  for(j=0;j<dga->Vsize+1;++j) {
    for(k=0;k<dga->Ysize;++k)
      wtVY[j][k] = wtvec[offset + dga->Ysize*j + k];
  }

  h[dga->Hsize]=HIDBIAS; /* initialises first hidden layer bias node */
  v[dga->Vsize]=HIDBIAS; /* initialises second hidden layer bias node */
  
  /* Evaluate total error for all patterns (batch update) */
  
  for(p=0;p<dga->Npats;++p) { /* patterns */
    if(dga->flag[14]>=4) printf("statmacfn(): p=%2d\n", p);
    if(dga->flag[53]==1)
      forpass(dga->Xsize, dga->Hsize, dga->Ysize, dga->x[p], h, y,
	      wtXH, wtVY, dga->Hlam, dga->flag);
    else if(dga->flag[53]==2)
      forpass2(dga->Xsize, dga->Hsize, dga->Vsize, dga->Ysize, 
	      dga->x[p], h, v, y, wtXH, wtHV, wtVY, dga->Hlam, dga->Vlam, dga->flag);
    else
      	neterr("statmacfn(): flag[53] not recognised");
    for(k=0;k<dga->Ysize;++k) { 
      /* only contribute to error if target defined */
      if(dga->tar[p][k].def) { 
	temp = (double)errfn(y[k],dga->tar[p][k].val);
	lerr += 0.5*dga->beta[k]*temp*temp;
      }
    }
    if(dga->flag[8]>0) /* weight decay contribution to error */
      werr += wderr(dga->Xsize, dga->Hsize, dga->Vsize, dga->Ysize, 
		    wtXH, wtHV, wtVY, dga->alpha, dga->flag); 
  } /* end of pattern loop */
  
  toterr=lerr+werr;
  if(dga->flag[14]>=3)
    printf("statmacfn(): Errors: L = %5.2e (%3.1f%%)  W = %5.2e (%3.1f%%)  Tot = %5.2e\n", lerr, 100*lerr/toterr, werr, 100*werr/toterr, toterr);

  /* deallocate memory */

  free_Dmatrix(wtXH, dga->Xsize+1, dga->Hsize);
  free_Dmatrix(wtHV, dga->Hsize+1, dga->Vsize);
  free_Dmatrix(wtVY, dga->Vsize+1, dga->Ysize);
  
  free_Dvector(h, dga->Hsize+1);
  free_Dvector(v, dga->Vsize+1);
  free_Dvector(y, dga->Ysize);

  /* decrement pointer to accommodate the unit offset in macopt */

  wtvec--;

  return(toterr);
  
}



/* ==================== Gradient Evaluation Routines =================== */

void ederiv(int Xsize, int Hsize, int Ysize, double *x, double *h, double *y, 
	    double **wtXH, double **wtHY, double **edXH, double **edHY,
	    targets *tar, double *alpha, double *beta, double Hlam, int *flag)
{
  /* Evaluate derivatives of error with respect to all weights (ij)
     for a single pattern (including weight decay contributions),
     and returns these: edXH and edHY. These are always initialised here. 

     NB This subroutine is only for a single hidden node. Hidden-Output
        weights/derivatives are called HY here, although they're passed
	in as VY.

     A forward pass of the network is done using forpass() before
     this routine is called. forpass() evaluates h and y
     which are passed into this routine.
     
     Note also that tar in this subroutine is a 1D vector of type
     targets rather than a 2D array of type targets as in the rest of
     the program.  
     */
  
  int k,l,m;
  double r1;        /* intermediate variables */
  double *s1,*s2;   /* intermediate sum variables 
		       - s1 could have been just a scalar */


  if(flag[14]>=4) printf("ederiv(): Evaluating error derivatives\n");

  /* allocate memory */

  s1 = Dvector(Ysize);
  s2 = Dvector(Hsize+1);

  /* initialise */

  for(m=0;m<Hsize+1;++m) /* s2[Hsize] calculated later but not used */
    s2[m]=0.0;

  for(m=0;m<Hsize;++m) {
    for(l=0;l<Xsize+1;++l)
      edXH[l][m]=0.0;
  }
  for(m=0;m<Hsize+1;++m) {
    for(k=0;k<Ysize;++k)
      edHY[m][k]=0.0;
  }

  /* evaluate edXH and edHY */

  for(k=0;k<Ysize;++k) {
    s1[k] = beta[k]*errfn(y[k],tar[k].val)*dTg(y[k]);
    for(m=0;m<Hsize+1;++m) { /* - including H bias */
      edHY[m][k] = s1[k]*h[m];
      s2[m] += s1[k]*wtHY[m][k];
    }
  }
  for(m=0;m<Hsize;++m) { /* - not including H bias */
    r1 = dTf(h[m],Hlam)*s2[m];
    for(l=0;l<Xsize+1;++l)
      edXH[l][m] = x[l]*r1;
  }

  /* Contributions from weight decay - note that if we're using weight
     decay then we get contributions (for non-zero alphas) even when
     outputs are not defined. Could this be a problem? Will we get too
     much regularization from weight decay error contributions? */
  
  if(flag[8]>0) { /* using weight decay */
    for(m=0;m<Hsize;++m) { /* not including H bias */
      for(l=0;l<Xsize;++l) /* XH */
	edXH[l][m]+=alpha[0]*wtXH[l][m];
      edXH[Xsize][m]+=alpha[1]*wtXH[Xsize][m]; /* input bias */
    }
    for(m=0;m<Hsize+1;++m) { /* including H bias */
      for(k=0;k<Ysize;++k) /* HY */
	edHY[m][k]+=alpha[3]*wtHY[m][k]; 
    }
  }  

  /* deallocate memory */

  free_Dvector(s1, Ysize);
  free_Dvector(s2, Hsize+1);

  if(flag[14]>=4) printf("ederiv(): Got to end\n");

}


void ederiv2(int Xsize, int Hsize, int Vsize, int Ysize, 
	     double *x, double *h, double *v, double *y, 
	     double **wtXH, double **wtHV, double **wtVY, 
	     double **edXH, double **edHV, double **edVY,
	     targets *tar, double *alpha, double *beta, 
	     double Hlam, double Vlam, int *flag)
{
  /* Evaluates derivatives of error with respect to all weights (ij)
     for a single pattern (including weight decay contributions) for
     TWO hidden layers, and returns edXH, edHV and edVY. 
     These are always initialised here. 

     A forward pass of the network is done using forpass2() before
     this routine is called. forpass2() evaluates h, v and y
     which are passed into this routine.
     
     Note also that tar in this subroutine is a 1D vector of type
     targets rather than a 2D array of type targets as in the rest of
     the program.  
     */
  
  int l,m,j,k;
  double r1,r2,r3;  /* intermediate variables */
  double *s1,*s2;   /* intermediate sum variables */


  if(flag[14]>=4) printf("ederiv2(): Evaluating error derivatives\n");

  /* allocate memory */

  s1 = Dvector(Vsize+1);
  s2 = Dvector(Hsize+1);

  /* initialise */

  for(j=0;j<Vsize+1;++j) /* s1[Vsize] calculated later but not used */
    s1[j]=0.0;
  for(m=0;m<Hsize+1;++m) /* s2[Hsize] calculated later but not used */
    s2[m]=0.0;
  for(l=0;l<Xsize+1;++l) {
    for(m=0;m<Hsize;++m)
      edXH[l][m]=0.0;
  }
  for(m=0;m<Hsize+1;++m) {
    for(j=0;j<Vsize;++j)
      edHV[m][j]=0.0;
  }
  for(j=0;j<Vsize+1;++j) {
    for(k=0;k<Ysize;++k)
      edVY[j][k]=0.0;
  }

  /* evaluate edXH, edHV and edVY */

  for(k=0;k<Ysize;++k) {
    r1 = beta[k]*errfn(y[k],tar[k].val)*dTg(y[k]);
    for(j=0;j<Vsize+1;++j) {
      edVY[j][k] = r1*v[j];
      s1[j] += r1*wtVY[j][k];
    }
  }
  for(j=0;j<Vsize;++j) {
    r2 = s1[j]*dTf(v[j],Vlam);
    for(m=0;m<Hsize+1;++m) {
      edHV[m][j] = r2*h[m];
      s2[m] += r2*wtHV[m][j];
    }
  }
  for(m=0;m<Hsize;++m) {
    r3 = s2[m]*dTf(h[m],Hlam);
    for(l=0;l<Xsize+1;++l)
      edXH[l][m] += r3*x[l];
  }

  /* Contributions from weight decay - note that if we're using weight
     decay then we get contributions (for non-zero alphas) even when
     outputs are not defined. Could this be a problem? Will we get too
     much regularization from weight decay error contributions? */
  
  if(flag[8]>0) { /* using weight decay */
    for(m=0;m<Hsize;++m) {
      for(l=0;l<Xsize;++l) /* XH */
	edXH[l][m]+=alpha[0]*wtXH[l][m];
      edXH[Xsize][m]+=alpha[1]*wtXH[Xsize][m]; /* input bias */
    }
    for(m=0;m<Hsize+1;++m) {
      for(j=0;j<Vsize;++j) /* HV */
	edHV[m][j]+=alpha[2]*wtHV[m][j];
    }
    for(j=0;j<Vsize+1;++j) {
      for(k=0;k<Ysize;++k) /* VY */
	edVY[j][k]+=alpha[3]*wtVY[j][k]; 
    }
  }  

  /* deallocate memory */

  free_Dvector(s1, Vsize+1);
  free_Dvector(s2, Hsize+1);

}




void cumederivs(int Xsize, int Hsize, int Vsize, int Ysize,
		double **edXH, double **edHV, double **edVY, 
		double **cumedXH, double **cumedHV, double **cumedVY, int *flag)
{
  /* cumulate the error derivatives (ed) into cumed 
     handles 1 and 2 hidden layers */

  int l,m,j,k;

  for(m=0;m<Hsize;++m) {
    for(l=0;l<Xsize+1;++l)
      cumedXH[l][m] += edXH[l][m];
  }
  if(flag[53]==2) {
    for(m=0;m<Hsize+1;++m) {
      for(j=0;j<Vsize;++j) 
	cumedHV[m][j] += edHV[m][j];
    }  
  }
  for(j=0;j<Vsize+1;++j) {
    for(k=0;k<Ysize;++k) 
      cumedVY[j][k] += edVY[j][k];
  }  

}



/* ==================== Initialisation Routines =================== */

void cominit(int Nnets, int Xsize, int Hsize, int Vsize, int Ysize,
	     double ***wtXH, double ***wtHV, double ***wtVY, 
	     double **yscale, double **xscale, double *Hlam, double *Vlam, 
	     int *flag, char *inwtfname, char *sfname, long ranseed,
	     double *alpha, double *beta, double wtrng)
{
  /* Initialise a committee of Nnets networks
     
     This routine is called if the committee has not just be trained,
     i.e. it is called by statnettrain and may be called by statnetapply.

     Hlam and Vlam are always set by this routine.
     Weights are either read in from weights file or randomized.
     Scaling factors are only set if read in from the weights file.
     
     */

  FILE *inwtf; /* input weight file */
  FILE *sf;    /* specfile */
  char message[MESSAGESIZE]; /* message */
  char word[WORDSIZE], tmp[WORDSIZE]; /* generic words */
  char scaleopt[WORDSIZE]; /* scaling option */
  int n,l,m,j,k;
  int Ntmp, Xtmp, Htmp, Vtmp, Ytmp;
  double betasum; /* used as a test of zero beta values */

  /* set Hlam and Vlam to default */

  *Hlam = 1.0;
  *Vlam = 1.0;
  
  if(flag[14]>=3)
    printf("cominit(): Network parameters (internal to statnet): %d %d %d %d %d\n", Nnets, Xsize, Hsize, Vsize, Ysize);

  if(flag[10]==1) { /* read weights and scaling factors from file */
    inwtf=fopen(inwtfname,"r");
    if(!inwtf) {
      sprintf(message, "cominit(): weights file %s could not be opened",
	      inwtfname);
      neterr(message);
    }
    if(flag[14]>=2)
      printf("cominit(): Reading weights from %s\n", inwtfname);

    fgets(word, WORDSIZE, inwtf); /* read line including \n */
    fgets(word, WORDSIZE, inwtf);
    fgets(word, WORDSIZE, inwtf);
    fgets(word, WORDSIZE, inwtf);
    sscanf(word, "%d %d %d %d %d\n", &Ntmp, &Xtmp, &Htmp, &Vtmp, &Ytmp);
    if(Ntmp!=Nnets || Xtmp!=Xsize || Htmp!= Hsize || Ytmp!=Ysize 
       || (flag[53]==1 && Vtmp!=0) || (flag[53]==2 && Vtmp!=Vsize))
      neterr("cominit(): Committee size/network size in weights file does not\n match spec file"); 

    fgets(word, WORDSIZE, inwtf);
    fgets(word, WORDSIZE, inwtf);
    sscanf(word,"%s",scaleopt);
    if(strcmp(scaleopt,"none")==0) {
      if(flag[12]!=0 && flag[14]>=1) {
	netwarn("cominit(): Input weights file indicates no scaling option in\n conflict with spec file. Former takes precedence and hence no\n scaling of input data will be performed");
	flag[12]=0;
      }
    }
    else if(strcmp(scaleopt,"var")==0) {
      if(flag[12]!=1 && flag[14]>=1) {
	netwarn("cominit(): Input weights file indicates variance scaling option\n in conflict with spec file. Former takes precedence and will be used\n to scale output variables and input data");
	flag[12]=1;
      }
      fgets(word, WORDSIZE, inwtf);
      for(k=0;k<Ysize;++k) {
	fgets(word, WORDSIZE, inwtf);
	sscanf(word, "%lf %lf", &yscale[0][k], &yscale[1][k]);
      }
      fgets(word, WORDSIZE, inwtf);
      for(l=0;l<Xsize;++l) { /* this does not count biases */
	fgets(word, WORDSIZE, inwtf);
	sscanf(word, "%lf %lf", &xscale[0][l], &xscale[1][l]);
      }
    }
    else if(strcmp(scaleopt,"maxmin")==0)
      neterr("cominit(): maxmin scaling option not yet implemented");
    else if(strcmp(scaleopt,"netsize")==0)
      ; /* don't do anything: Hlam (and Vlam) is read in for all options anyway */
    else 
      neterr("cominit(): invalid scaling option in weights file");
    
    fgets(word, WORDSIZE, inwtf);
    fgets(word, WORDSIZE, inwtf);
    sscanf(word, "%lf", Hlam);
    if(flag[53]==2) {
      fgets(word, WORDSIZE, inwtf);
      fgets(word, WORDSIZE, inwtf);
      sscanf(word, "%lf", Vlam);
    }

    /* read in weights for each network in committee in turn */

    for(n=0;n<Nnets;++n) {
      fgets(word, WORDSIZE, inwtf);
      fgets(word, WORDSIZE, inwtf);
      for(l=0;l<Xsize+1;++l) {
	for(m=0;m<Hsize;++m)
	  fscanf(inwtf, "%lf", &wtXH[n][l][m]);
	fgets(word, WORDSIZE, inwtf); /* to remove \n */
      }
      if(flag[53]==2) {
	fgets(word, WORDSIZE, inwtf);
	for(m=0;m<Hsize+1;++m) {
	  for(j=0;j<Vsize;++j)
	    fscanf(inwtf, "%lf", &wtHV[n][m][j]);
	  fgets(word, WORDSIZE, inwtf);
	}
      }
      fgets(word, WORDSIZE, inwtf);
      for(j=0;j<Vsize+1;++j) {
	for(k=0;k<Ysize;++k)
	  fscanf(inwtf, "%lf", &wtVY[n][j][k]);
	fgets(word, WORDSIZE, inwtf);
      }
    } /* end network */

    fclose(inwtf);
  }

  else if(flag[10]==0) { /* initialise weights
			    NB scaling factors not set here */
    if(flag[14]>=2)
      printf("cominit(): Initialising weights\n");
    if(flag[9]==1) { /* uniform weight distribution */
      for(n=0;n<Nnets;++n) {
	for(l=0;l<Xsize+1;++l) {
	  for(m=0;m<Hsize;++m)
	    wtXH[n][l][m] = 2*wtrng*(double)(ran1(&ranseed)-0.5);
	}
	if(flag[53]==2) {
	  for(m=0;m<Hsize+1;++m) {
	    for(j=0;j<Vsize;++j)
	    wtHV[n][m][j] = 2*wtrng*(double)(ran1(&ranseed)-0.5);
	  }
	}
	for(j=0;j<Vsize+1;++j) {
	  for(k=0;k<Ysize;++k)
	    wtVY[n][j][k] = 2*wtrng*(double)(ran1(&ranseed)-0.5);
	}
      }
    }
    else
      neterr("cominit(): Only a uniform weight distribution\n for initialisation is currently available\n Bummer huh?");
  }
  
  /* set alpha and beta parameters only if training
   - searches specfile twice. Slow maybe, but makes neater code. 
   would be better to have this as a function separate from cominit() ?
   */

  if(flag[1]==0)
    return;

  /* set alpha parameters */

  if(flag[8]==2) {
    sf=fopen(sfname,"r");
    if(!sf) {
      sprintf(message, "cominit(): specfile %s could not be opened",
	      sfname);
      neterr(message);
    }
    while(fscanf(sf,"%s",word)!=EOF) {
      if(word[0]=='#') { /* NB Must use single quotes! */
	fgets(word, WORDSIZE, sf); /* read rest of line, including \n */
	continue;
      }
      if(strcmp(word,"TRN:alpha_XH")==0) {
	fscanf(sf,"%lf",&alpha[0]);
	if(alpha[0]<0.0)
	  neterr("cominit(): alpha_XH is an invalid negative value");
      }
      if(strcmp(word,"TRN:alpha_bH")==0) {
	fscanf(sf,"%lf",&alpha[1]);
	if(alpha[1]<0.0) 
	  neterr("cominit(): alpha_bH is an invalid negative value");
      }
      if(strcmp(word,"TRN:alpha_HV")==0) {
	fscanf(sf,"%lf",&alpha[2]);
	if(alpha[2]<0.0)
	  neterr("cominit(): alpha_HV is an invalid negative value");
      }
      if(strcmp(word,"TRN:alpha_VY")==0) {
	fscanf(sf,"%lf",&alpha[3]);
	if(alpha[3]<0.0)
	  neterr("cominit(): alpha_VY is an invalid negative value");
      }
    } /* end of file searching while loop */
    fclose(sf);
  }
  else if(flag[8]==1) {
    alpha[0]=alpha_XH_DEF;
    alpha[1]=alpha_bH_DEF;
    alpha[2]=alpha_HV_DEF;
    alpha[3]=alpha_VY_DEF;
  }
  else { /* strictly unnecessary: just for security */
    for(n=0;n<NALPHA;++n)
      alpha[n]=0.0;
  }
  
  /* set beta parameters
  
     Note that if we do not set a sufficient number of values, then
     the last value is used for all the values of beta where a value
     is not specified. However, it's best to assign your own
     values. */
     
  if(flag[7]==2) {
    sf=fopen(sfname,"r");
    if(!sf) {
      sprintf(message, "cominit(): specfile %s could not be opened",
	      sfname);
      neterr(message);
    }
    while(fscanf(sf,"%s",word)!=EOF) {
      if(word[0]=='#') { /* NB Must use single quotes! */
	fgets(word, WORDSIZE, sf); /* read rest of line, including \n */
	continue;
      }
      if(strcmp(word,"TRN:use_beta_parameters?_(none/default/list)")==0) {
	fgets(word, WORDSIZE, sf); /* to remove word */
	for(k=0;k<Ysize;++k) {
	  fgets(word, WORDSIZE, sf); 
	  sscanf(word,"%s %lf", tmp, &beta[k]);  
	  /* ignore value if incorrect stem */ 
	  if(strcmp(tmp,BETASTEM)!=0)
	    beta[k]=BETADEF;
	}
	break; /* stop scaning specfile */
      }
    }
    fclose(sf);
    for(k=1;k<Ysize;++k) { /* if no beta specified for any k, assign
			      last non-zero value - doesn't work unless
			      at least beta[0] is assigned */
      if(beta[k]==0.0)
	beta[k]=beta[k-1];
    }
  }
  else if(flag[7]==1) { /* use defaults */
    for(k=0;k<Ysize;++k)
      beta[k]=BETADEF;
  }
  else { /* don't use beta - i.e. set to 1.0 */
    for(k=0;k<Ysize;++k)
      beta[k]=1.0;
  }

  /* print out and check values of beta */

  betasum=0.0;
  for(k=0;k<Ysize;++k) {
    if(flag[14]>=2) printf("cominit(): beta[%d] = %9.5f\n", k, beta[k]);
    if(beta[k]<0.0)
      neterr("cominit(): Invalid negative beta value");
    if(beta[k]==0.0 && flag[14]>=1) {
      sprintf(message,"cominit(): output %d has a zero beta value.\n This output variable will not contribute to the error function,\n so it will not be constrained by the training", k+1);
      netwarn(message);      
    }
    betasum+=beta[k];
  }
  if(flag[7]>0 && betasum==0.0 && flag[14]>=1)
    netwarn("cominit(): All beta values are zero. Training will be data independent");


}




/* ==================== Scaling Routines =================== */

void scalecalc(int Xsize, int Hsize, int Ysize, int Npats, 
	       double **x, targets **tar, double **xscale, double **yscale, 
	       double *Hlam, double *Vlam, int *flag)
{
  /* Calculate Hlam, Vlam and the data scale factors for inputs and ouputs

     If there are no defined values for a given node, scaling should
     not affect what occurs in rest of program. Thus:
       mean-sd scaling -> mean=0, sd=1
     If there is only one defined value on a given node or all 
     defined values are equal:
       mean-sd scaling -> mean=<value>, sd=1
     These prevent divide by zero in datascale()

     */

  
  int p,k,l;
  int numvecs;
  int *ynum;
  long double tmp; 
  long double ytmp;
  long double *xtmp;
  char message[MESSAGESIZE];
  

  /* evaluate Hlam */
  
  *Hlam = 1/( sqrt((double)(Xsize+1)) );
  if(flag[53]==2)
    *Vlam = 1/( sqrt((double)(Hsize+1)) );

  if(flag[12]==3) /* i.e. only set Hlam and Vlam */
    return;

  if(flag[12]!=1)
    neterr("scalecalc(): Scaling option not yet implemented");
  /* put all of the following into an if clause if and when you write
     the minmax scaling option. Also, if you implement on-line learning
     you could evaluate some scale factors here */

  if(flag[14]>=2)
    printf("scalecalc(): Calculating data scale\n");
  
  /* initialise xscale */

  xtmp = lDvector(Xsize);
  for(l=0;l<Xsize;++l) {
    xscale[0][l]=0.0;
    xscale[1][l]=0.0;
    xtmp[l]=0.0;
  }

  /* first pass: evaluate mean */
  for(p=0,numvecs=0;p<Npats;++p) { /* patterns */
    numvecs++;
    for(l=0;l<Xsize;++l)  
      xtmp[l] += (long double)x[p][l];
    if(flag[14]>=4)
      printf("mean for input no.1: %3d % 8.5e % 8.5e\n", p, x[p][0], (double)xtmp[0]);
  }
  for(l=0;l<Xsize;++l) {
    xscale[0][l] = (xtmp[l]/(long double)numvecs);
    xtmp[l] = 0.0; /* re-initialise */
  }

  /* second pass: evaluate standard deviation */
  for(p=0;p<Npats;++p) { /* patterns */
    for(l=0;l<Xsize;++l) { 
      tmp = (long double)x[p][l] - (long double)xscale[0][l];
      xtmp[l] += tmp*tmp; 
    }
    if(flag[14]>=4)
      printf("stdev for input no.1: %3d % 8.5e % 8.5e\n", p, x[p][0], (double)xtmp[0]);
  }

  for(l=0;l<Xsize;++l) {
    xscale[1][l] = (double)(sqrt( (double)(xtmp[l]/(long double)numvecs) ));
    if(flag[14]>=2)
      printf(" input  (x) %2d: mean = % 8.5e  sd = % 8.5e\n", l+1, xscale[0][l], xscale[1][l]);
  }
  for(l=0;l<Xsize;++l) {
    if(xscale[1][l]==0.0) { /* i.e. all inputs for a given node are equal */
      xscale[1][l]=1.0;
      if(flag[14]>=1) {
	sprintf(message, "scalecalc(): X standard deviation for l=%d is zero (all inputs the same?)\n Forcing stddev=1 to prevent divide by zero", l);
	netwarn(message);
      }
    }
  }

  free_lDvector(xtmp, Xsize);

  /* initialise yscale - here the number of nodes is the outermost loop,
     in contrast to the xscale calculation, where the number of patterns
     was the outermost loop. There was probably a good reason for this! */
  
  ynum=Ivector(Ysize);
  for(k=0;k<Ysize;++k) {
    yscale[0][k]=0.0;
    yscale[1][k]=0.0;
    ynum[k]=0;
  }

  /* first pass: evaluate mean */
  for(k=0;k<Ysize;++k) { /* nodes */
    ytmp=0.0;
    for(p=0,numvecs=0;p<Npats;++p) { /* patterns */
      if(tar[p][k].def) {
	ytmp += (long double)tar[p][k].val;
	ynum[k]++;
      }
    } /* end of patterns */
    if(ynum[k]==0)
      yscale[0][k]=0.0;
    else
      yscale[0][k] = (double)(ytmp/(long double)ynum[k]);
  }
  
  /* second pass: evaluate standard deviation */
  for(k=0;k<Ysize;++k) { /* nodes */
    if(ynum[k]<=1)
      yscale[1][k]=1.0; /* Force sd to unity
			   to prevent divide by zero in datascale() */
    else {
      ytmp=0.0;
      for(p=0;p<Npats;++p) { /* patterns */
	if(tar[p][k].def) {
	  tmp   = (long double)tar[p][k].val - (long double)yscale[0][k];
	  ytmp += tmp*tmp; 
	}
      } /* end of patterns */
      yscale[1][k] = (double)(sqrt( (double)(ytmp/(long double)ynum[k]) ));
      if(flag[14]>=2)
	printf(" output (y) %2d: mean = % 8.5e  sd = % 8.5e\n", k+1, yscale[0][k], yscale[1][k]);
    }
  } /* end of nodes */

  for(k=0;k<Ysize;++k) {
    if(yscale[1][k]==0.0) {
      yscale[1][k]=1.0;
      if(flag[14]>=1) {
	sprintf(message, "datascale(): Y standard deviation for k=%d is zero (all targets the same?)\n Forcing stddev=1 to prevent divide by zero", k);
	netwarn(message);
      }
    }
  }
  
  free_Ivector(ynum, Ysize);
    
}



void datascale(int Xsize, int Ysize, int Npats, double **x,
	       targets **tar, double **xscale, double **yscale, int *flag)
{
  /* scale the inputs and outputs using known scaling parameters */


  int p,k,l;
  char message[MESSAGESIZE];

  
  if(flag[14]>=2) printf("datascale(): Scaling data\n");

  if(flag[12]==1) { /* mean and standard deviation scale */

    for(p=0;p<Npats;++p) { /* patterns */
      if(flag[14]>=3) printf("p = %2d\n", p);
      /* scale y */
      if(flag[14]>=3) printf("targets (scaled values):\n");
      for(k=0;k<Ysize;++k) { /* nodes */
	if(tar[p][k].def) {
	  if(yscale[1][k]==0.0) {
	    sprintf(message, "datascale(): attempt to divide by zero (Y standard deviation for k=%d is zero)", k);
	    neterr(message);
	  }
	  tar[p][k].val = (tar[p][k].val-yscale[0][k])/yscale[1][k];
	  if(flag[14]>=3) printf("% 9.5f ", tar[p][k].val);
	}
	else {
	  if(flag[14]>=3) printf("  x       ");
	}
      } /* end of nodes */
      if(flag[14]>=3) printf("\n"); 

      /* scale x */
      if(flag[14]>=3) printf("external inputs (scaled values):\n");
      for(l=0;l<Xsize;++l) {
	if(xscale[1][l]==0.0) {
	  sprintf(message, "datascale(): attempt to divide by zero (X standard deviation for l=%d is zero)", l);
	  neterr(message);
	}
	x[p][l] = (x[p][l]-xscale[0][l])/xscale[1][l];
	if(flag[14]>=3) printf("% 9.5f ", x[p][l]);
      }
      if(flag[14]>=3) printf("\n"); 
    } /* end of patterns */

  }
  else
    neterr("datascale(): Scaling option not yet implemented");
  
}



void unscale(int Ysize, double *y, double **yscale, int *flag)
{
  /* unscale a vector */

  int k;
  
  if(flag[12]==1) { /* mean and standard deviation scale */
    for(k=0;k<Ysize;++k)
      y[k] = y[k]*yscale[1][k] + yscale[0][k];
  }
  else
    neterr("unscale(): Scaling option not yet implemented");
}



/* ==================== Input/Output Routines =================== */

void specread(char *sfname, int *flag, 
	      int *Nnets, int *Xsize, int *Hsize, int *Vsize, int *Ysize, 
	      int *Xtot, int *Ytot,
	      int **xranges, int **yranges, int *Nits, double *eta,
	      char *trnfname, char *appfname,
	      char *inwtfname, char *otwtfname, long *ranseed,
	      double *macconvtol, int *macitmax, double *macchecktol,
	      double *wtrng, char *errfname, int *takelog, char *exclude[])
{
  /* Read in spec file 

     The general rule is that if a bogus input is specified, flags
     etc. will be set to their default values without warning.
     Exceptions are where inputs are NOT simply "yes/no".

     Need to pass back read-in variables, so defined them in as pointers 
     */
  
  FILE *sf;
  char message[MESSAGESIZE];
  char word[WORDSIZE];
  char word2[WORDSIZE];
  double fnum;
  int inum,r,nlog,p;

  /* set specifications (flags etc.) to their default values. Some 
     values are bizarre to serve as checks on whether they've been set
  */

  flag[1] = 0;
  flag[2] = 0;  
  flag[3] = 2;
  flag[4] = 3;
  flag[7] = 0;
  flag[8] = 0;
  flag[9] = 1;
  flag[10]= 0;
  flag[11]= 0;
  flag[12]= 1;
  flag[13]= 0;
  flag[14]= 2;
  flag[16]= 0;
  flag[17]= 0;
  flag[18]= 0;
  *Xtot=-1;
  *Ytot=-1;
  *Xsize=-1;
  *Ysize=-1;
  xranges[0][0]=0;
  yranges[0][0]=0;
  *Nnets=1;
  *Hsize=-1;
  *Vsize=-1;
  *wtrng=WTRNGDEF;
  *ranseed=RANSEEDDEF;
  *Nits=ITSDEF;
  *eta=ETADEF;
  *macconvtol=MACCONVTOLDEF;
  *macitmax=ITSDEF;
  *macchecktol=MACCHECKTOLDEF;
  /* internal macopt parameters are set to default values by calling
     macopt_defaults() in callmacopt() */
  
  /* read specfile */
  
  sf=fopen(sfname,"r");
  if(!sf) {
    sprintf(message, "specread(): specfile %s could not be opened", sfname);
    neterr(message);
  }

  while(fscanf(sf,"%s",word)!=EOF) {

    if(word[0]=='#') { /* NB Must use single quotes! */
      fgets(word, WORDSIZE, sf); /* read rest of line, including \n */
      continue;
    }
    else if(strcmp(word,"verbosity_level_(0/1/2/3/4)")==0) {
      fscanf(sf,"%d",&inum);
      if(inum>=0 && inum<=4)
	flag[14]=inum; /* otherwise retain default */
    }
    else if(strcmp(word,"train_network?_(yes/no)")==0) {
      fscanf(sf,"%s",word);
      if(strcmp(word,"yes")==0) flag[1]=1;
    }
    else if(strcmp(word,"apply_network?_(yes/no)")==0) {
      fscanf(sf,"%s",word);
      if(strcmp(word,"yes")==0) flag[2]=1;
    }
    else if(strcmp(word,"DAT:training_data_file")==0) {
      fscanf(sf,"%s",trnfname);
    }
    else if(strcmp(word,"DAT:application_data_file")==0) {
      fscanf(sf,"%s",appfname);
    }
    else if(strcmp(word,"DAT:number_of_inputs_in_data_file_(X)")==0) {
      fscanf(sf,"%d",&inum);
      if(inum<0)
	neterr("specread(): Number of inputs (X) cannot be negative");
      if(inum>NODEWARNSIZE && flag[14]>=1)
	netwarn("specread(): That's a lot of inputs (X)");
      *Xtot=inum;
    }
    if(strcmp(word,"DAT:number_of_input_ranges")==0) {
      if(*Xtot==-1)
	neterr("specread(): X must be defined before input ranges in spec file");
      fscanf(sf,"%d",&xranges[0][0]);
      if(xranges[0][0]>0 && xranges[0][0]<=*Xtot) { /* read ranges */
	*Xsize=0;
	for(r=1;r<=xranges[0][0];++r) { /* loop over ranges */
	  do { /* ensure that prefix string is INRANGESTEM */
	    inum=fscanf(sf,"%s",word);
	    fgets(word2, WORDSIZE, sf); /* read rest of line, including \n */
	    if(inum==EOF)
	      neterr("specread(): end of file reached before all input data ranges found\n");
	  }
	  while (strcmp(word,INRANGESTEM)!=0);
	  sscanf(word2,"%d %d",&xranges[0][r],&xranges[1][r]);
	  if(flag[14]>=3) printf("input  range %d: lower = %3d, upper = %3d\n", r, xranges[0][r], xranges[1][r]);
	  /* trap errors */
	  if(xranges[1][r]<xranges[0][r]) {
	    sprintf(message, "specread(): lower input range value is larger than upper value in range number %d", r);
	    neterr(message);
	  }
	  else if(xranges[1][r]>*Xtot) {
	    sprintf(message, "specread(): upper value in input range %d exceeds\n number of inputs in data file (X=%d)", r, *Xtot);
	    neterr(message);
	  }
	  else if(xranges[0][r]<1) {
	    sprintf(message, "specread(): lower value in input range %d is < 1", r);
	    neterr(message);
	  }
	  *Xsize+=xranges[1][r]-xranges[0][r]+1;
	} /* end ranges */
      }
      else if(xranges[0][0]==0)
	;  /* dummy value for selecting entire range dealt with at end */
      else
	neterr("specread(): Number of input ranges (xranges[0][0]) is illogical");
    }
    else if(strcmp(word,"DAT:number_of_outputs_in_data_file_(Y)")==0) {
      fscanf(sf,"%d",&inum);
      if(inum<0)
	neterr("specread(): Number of output (Y) cannot be negative");
      if(inum>NODEWARNSIZE && flag[14]>=1)
	netwarn("specread(): That's a lot of outputs (Y)");
      *Ytot=inum;
      if(*Ytot>MAXOUTS) neterr("Increase MAXOUTS in statnet.h to be at least as large as number of outputs\n");
    }
    if(strcmp(word,"DAT:number_of_output_ranges")==0) {
      if(*Ytot==-1)
	neterr("specread(): Y must be defined before output ranges in spec file");
      fscanf(sf,"%d",&yranges[0][0]);
      if(yranges[0][0]>0 && yranges[0][0]<=*Ytot) { /* read ranges */
	*Ysize=0;
	for(r=1;r<=yranges[0][0];++r) { /* loop over ranges */
	  do { /* ensure that prefix string is OTRANGESTEM, 
		  e.g. ignore comment lines */
	    inum=fscanf(sf,"%s",word);
	    fgets(word2, WORDSIZE, sf); /* read rest of line, including \n */
	    if(inum==EOF)
	      neterr("specread(): end of file reached before all output data ranges found\n");
	  }
	  while (strcmp(word,OTRANGESTEM)!=0);
	  sscanf(word2,"%d %d",&yranges[0][r],&yranges[1][r]);	  
	  if(flag[14]>=3) printf("output range %d: lower = %3d, upper = %3d\n", r, yranges[0][r], yranges[1][r]);
	  /* trap errors */
	  if(yranges[1][r]<yranges[0][r]) {
	    sprintf(message, "specread(): lower output range value is larger than upper value in range number %d", r);
	    neterr(message);
	  }
	  else if(yranges[1][r]>*Ytot) {
	    sprintf(message, "specread(): upper value in output range %d exceeds\n number of outputs in data file (Y=%d)", r, *Ytot);
	    neterr(message);
	  }
	  else if(yranges[0][r]<1) {
	    sprintf(message, "specread(): lower value in output range %d is < 1", r);
	    neterr(message);
	  }
	  *Ysize+=yranges[1][r]-yranges[0][r]+1;
	} /* end ranges */
      }
      else if(yranges[0][0]==0)
	;  /* dummy value for selecting entire range dealt with at end */
      else
	neterr("specread(): Number of output ranges (yranges[0][0]) is illogical");
    }
    else if(strcmp(word,"DAT:number_of_patterns_to_exclude")==0) {
      fscanf(sf,"%d",&inum);
      if(inum<0)
	neterr("specread(): Number of patterns to exclude cannot be negative");
      flag[18]=inum;
      for(p=0;p<flag[18];++p) {
	fscanf(sf,"%s",word);
	if(word[0]=='#') { /* this accommodates hashed out files */
	  fgets(word, WORDSIZE, sf); /* read rest of line, including \n */
	  p--;
	  continue;
	}
	exclude[p]=Cvector((int)(strlen(word)+1)); /* deallocated in main() */
	strcpy(exclude[p],word);
      }
    }
    else if(strcmp(word,"NET:number_of_networks_in_committee_(N)")==0) {
      fscanf(sf,"%d",&inum);
      if(inum<=0)
	neterr("specread(): Number of networks (N) must be positive");
      if(inum>COMWARNSIZE && flag[14]>=1)
	netwarn("specread(): That's a lot of networks (N)");
      *Nnets=inum;
    }
    else if(strcmp(word,"NET:size_of_first_hidden_layer_(H)")==0) {
      fscanf(sf,"%d",&inum);
      if(inum<0)
	neterr("specread(): Number of hidden nodes (H) cannot be negative");
      if(inum>NODEWARNSIZE && flag[14]>=1)
	netwarn("specread(): That's a lot of hidden nodes (H)");
      *Hsize=inum;
    }
    else if(strcmp(word,"NET:size_of_second_hidden_layer_(V)")==0) {
      fscanf(sf,"%d",&inum);
      if(inum<0)
	neterr("specread(): Number of hidden nodes (V) cannot be negative");
      if(inum>NODEWARNSIZE && flag[14]>=1)
	netwarn("specread(): That's a lot of hidden nodes (V)");
      *Vsize=inum;
    }
    else if(strcmp(word,"NET:data_scaling_(none/var/maxmin/netsize)")==0) {
      fscanf(sf,"%s",word);
      if(strcmp(word,"none")==0)
	flag[12]=0;
      else if(strcmp(word,"var")==0)
	flag[12]=1;
      else if(strcmp(word,"maxmin")==0)
	flag[12]=2;
      else if(strcmp(word,"netsize")==0)
	flag[12]=3;
      else
	neterr("specread(): Option for\nNET:data_scaling_method_(none/var/maxmin) is not valid");
    }
    else if(strcmp(word,"NET:input_weight_file")==0) {
      fscanf(sf,"%s",inwtfname);
      flag[10]=1;
    }
    else if(strcmp(word,"NET:number_of_outputs_to_log")==0) {
      fscanf(sf,"%d",&nlog);
      if(nlog<0)
	neterr("specread(): NET:number_of_outputs_to_log must be >=0");
      if(*Ytot==-1) neterr("specread(): Y must be defined before list of variables to log in spec file");
      for(r=1;r<=nlog;r++) {
	do { /* ensure that prefix string is LOGSTEM */
	  if(EOF==fscanf(sf,"%s",word)) neterr("specread(): end of file reached before all variables to log identified\n");
	  fgets(word2, WORDSIZE, sf); /* read rest of line, including \n */
	}
	while (strcmp(word,LOGSTEM)!=0);
	sscanf(word2,"%d",&inum);
	takelog[inum]=-1; /* set it to this for now; set to value in range
			     in dataread() */
      }
    }
    else if(strcmp(word,"TRN:optimization_method_(grd/macopt)")==0) {
      fscanf(sf,"%s",word);
      if(strcmp(word,"macopt")==0) flag[3]=2;
      else if(strcmp(word,"grd")==0) flag[3]=1;
      else neterr("specread(): Option for\nTRN:Optimization_method_(grd/macopt) is not valid");
    }
    else if(strcmp(word,"TRN:update_method_(svu/rsu/batch)")==0) {
      fscanf(sf,"%s",word);
      if(strcmp(word,"svu")==0)
	flag[4]=1;
      else if(strcmp(word,"rsu")==0)
	flag[4]=2;
      else if(strcmp(word,"batch")==0)
	flag[4]=3;
      else
	neterr("specread(): Option for TRN:update_method_(svu/rsu/batch) is not valid");
    }
    else if(strcmp(word,"TRN:weight_decay_(none/default/list)")==0) {
      fscanf(sf,"%s",word);
      if(strcmp(word,"none")==0)
	flag[8]=0;
      else if(strcmp(word,"default")==0)
	flag[8]=1;
      else if(strcmp(word,"list")==0)
	flag[8]=2;
      else
	neterr("specread(): Option for TRN:weight_decay_(none/default/list) is not valid");
      /* alpha parameters themselves are read in later in statnettrain */
    }
    else if(strcmp(word,"TRN:use_beta_parameters?_(none/default/list)")==0) {
      fscanf(sf,"%s",word);
      if(strcmp(word,"none")==0)
	flag[7]=0;
      else if(strcmp(word,"default")==0)
	flag[7]=1;
      else if(strcmp(word,"list")==0)
	flag[7]=2;
      else
	neterr("specread(): Option for TRN:use_beta_parameters?_(none/default/list) is not valid");
      /* beta parameters themselves are read in later in statnettrain */
    }
    else if(strcmp(word,"TRN:output_weight_file")==0) {
      fscanf(sf,"%s",otwtfname);
      flag[11]=1;
    }
    else if(strcmp(word,"TRN:error_file")==0) {
      fscanf(sf,"%s",errfname);
      flag[16]=1;
    }
    else if(strcmp(word,"TRN:form_of_weight_init_(uniform/gaussian)")==0) {
      fscanf(sf,"%s",word);
      if(strcmp(word,"uniform")==0)
	flag[9]=1;
      else if(strcmp(word,"gaussian")==0)
	flag[9]=2;
      else
	neterr("specread(): Option for\nTRN:Form_of_weight_init_(uniform/gaussian) ... is not valid");
    }
    else if(strcmp(word,"TRN:initial_weight_range")==0) {
      fscanf(sf,"%lf",wtrng);
      if(fabs(*wtrng)>WTRNGWARNSIZE && flag[14]>=1)
	netwarn("specread(): Initial weight range is very large");
    }
    else if(strcmp(word,"TRN:random_number_seed")==0) {
      fscanf(sf,"%d",&inum);
      *ranseed=-(long)(abs((int)inum));
    }
    else if(strcmp(word,"GRD:number_of_iterations")==0) {
      fscanf(sf,"%d",&inum);
      if(inum<=0)
	neterr("specread(): Number of iterations must exceed zero");
      *Nits=inum;    
    }
    else if(strcmp(word,"GRD:learning_rate")==0) {
      fscanf(sf,"%lf",&fnum);
      if(fnum<=0.0)
	neterr("specread(): Learning rate must exceed zero");
      *eta=fnum;    
    }
    else if(strcmp(word,"MAC:convergence_tolerance_gradient")==0) {
      fscanf(sf,"%lf",macconvtol); 
      *macconvtol *= *macconvtol;
      if(*macconvtol<=0.0)
	neterr("specread(): MAC:convergence_tolerance_gradient must be positive");
    }
    else if(strcmp(word,"MAC:maximum_number_of_iterations")==0) {
      fscanf(sf,"%d",macitmax); 
      if(*macitmax<0)
	neterr("specread(): MAC:maximum_number_of_iterations cannot be negative");
    }
    else if(strcmp(word,"MAC:perform_maccheckgrad?_(yes/no)")==0) {
      fscanf(sf,"%s",word);
      if(strcmp(word,"yes")==0)
	flag[13]=1;
    }
    else if(strcmp(word,"MAC:maccheckgrad_tolerance")==0) {
      fscanf(sf,"%lf",macchecktol); 
      if(*macchecktol<=0.0)
	neterr("specread(): MAC:maccheckgrad_tolerance must be positive");
    }
    else if(strcmp(word,"APP:write_individual_network_results?_(yes/no)")==0) {
      fscanf(sf,"%s",word);
      if(strcmp(word,"yes")==0) flag[17]=1;
    }

  }
  fclose(sf);

  /* use all X and Y data if no ranges were given */

  if(xranges[0][0]==0) {
    xranges[0][0]=1;
    xranges[0][1]=1;
    xranges[1][1]=*Xtot;
    *Xsize=*Xtot;
  }
  if(yranges[0][0]==0) {
    yranges[0][0]=1;
    yranges[0][1]=1;
    yranges[1][1]=*Ytot;
    *Ysize=*Ytot;
  }

  /* check for errors - will only report first one */

  if(*Xtot==-1)  neterr("specread(): number of inputs in file has not been set");
  if(*Ytot==-1)  neterr("specread(): number of outputs in file has not been set");
  if(*Xsize==-1) neterr("specread(): number of input nodes has not been set");
  if(*Hsize==-1) neterr("specread(): size of first hidden layer has not been set");
  if(*Vsize==-1) neterr("specread(): size of second hidden layer has not been set");
  if(*Ysize==-1) neterr("specread(): number of output nodes has not been set");
  if(*Xsize>*Xtot) neterr("specread(): Sum of input ranges exceeds stated number of inputs in data file\n Have you specified overlapping ranges?");
  if(*Ysize>*Ytot) neterr("specread(): Sum of output ranges exceeds stated number of outputs in data file\n Have you specified overlapping ranges?");
  if(flag[4]==2) neterr("specread(): RSU (random subset update) training not yet implemented");
  if(flag[12]==2) neterr("specread(): maxmin scaling option not yet implemented");
  if(flag[9]==2) neterr("specread(): gaussian initial weights option not yet implemented");

  /* check consistency of options/flags and give warnings as appropriate */

  if(*Vsize==0) {
    flag[53]=1;
    *Vsize=*Hsize;
  }
  else
    flag[53]=2;

  if(flag[1]==0 && flag[10]==0 && flag[2]==1 && flag[14]>=1) {
    netwarn("specread(): Application of network without training or reading of weights\n file requested. Application will take place with random weights.");
    if(flag[12]==1 || flag[12]==2)
      netwarn("specread(): Scaling has been requested with random weights. Scale factors \n will be evaluated using the specified application file.");
  }
  if(flag[1]==1 && flag[10]==1 && flag[14]>=1) 
    netwarn("specread(): Training of network starting from given weights requested\n New data will be scaled according to scales in given weights file\n This will over-ride scaling method requested in spec file if necessary.");
  if(flag[3]==2 && flag[4]!=3) {
    flag[4]=3;
    if(flag[14]>=1) netwarn("specread(): Only batch update possible with macopt; setting flag[4]=3");
  }
  if(flag[12]==0 && flag[1]==1)
    netwarn("specread(): No data scaling requested. Weight saturation may occur");
  if(flag[12]==3 && flag[1]==1)
    netwarn("specread(): Only netsize (Hlam/Vlam) data scaling requested\n Weight saturation may occur");
  if(flag[14]==0 && flag[13]!=0) /* don't run maccheckgrad at zero verbosity */
    flag[13]=0;


  /* don't overwrite existing weights files; use default weight file if
     none specified. (Note that if we apply the network with random weights
     the existing weights file will be overwritten without warning) */

  if(flag[1]==1 && flag[11]==1 && (sf=fopen(otwtfname,"r")) ) {
    fclose(sf);
    sprintf(message, "specread(): weights file %s already exists\n Kill program now to prevent overwrite by writeweights()", otwtfname);
    netwarn(message);
  }
  if(flag[1]==1 && flag[11]==0) {
    strcpy(otwtfname,DEFWTNAME);
    flag[11]=1;
    if(flag[14]>=1) {
      sprintf(message, "specread(): Network training requested but no output weights file specified.\n Weights will be written to file %s", DEFWTNAME);
      netwarn(message);
    }
  }

  /* confirm architecture */

  if(flag[14]>=2) {
    if(flag[53]==1)
      printf("Selected network architecture (excluding biases) is %d:%d:%d\n", *Xsize,*Hsize,*Ysize);
    else
      printf("Selected network architecture (excluding biases) is %d:%d:%d:%d\n", 
	     *Xsize,*Hsize,*Vsize,*Ysize);
  }

}



void dataread(int Xsize, int Ysize, int Xtot, int Ytot, 
	      int **xranges, int **yranges,
	      int *Npats, char *datfname, double ***x, targets ***tar, 
	      int *takelog, int *flag, char **exclude, int *Nexcluded)
{
  /*  Read in training/application data file

      tar is the name for the ideal values of outputs (targets).  When
      applying (rather than training) the network, they are the ideal
      values which we may have in a file (as required when testing the
      network).

      Note that we pass in pointers to **x and pointes to **tar.
      See comments at top of statnettrain() subroutine, just after x and 
      tar are defined.

      In the bpnn codes, the format of the data files was <inputs> <outputs>
      on each line. THIS IS NOW REVERSED: <outputs> <inputs>. 

      Note that the xranges and yranges must be arranged in increasing
      order and may not overlap. Failure to do this shouldn't result
      in a crash (I think I've trapped those errors) but could result
      in unexpected behaviour.

      As from v1.02, Npats is modified after being read from he file
      header to take account of excluded patterns. However, as dataread()
      is one of the first functions called, this shouldn't affect the
      rest of the program. The one exception is the memory allocation
      of x and tar, which take place before Npats was reduced: Nexcluded
      is passed back for this purpose. This could have been done more
      neatly by deallocating the extra unrequired memory within this
      function.
      */
      
  int k,l; /* counts over the output/input variables which have been selected
	      i.e. are in yranges/xranges */
  int f;   /* counts the field in the input line, with f=1 as the first output
	      variable or the first input variable */
  int r;   /* counts over yranges/xranges */
  int p;   /* counts over the patterns */
  int e;    /* counts over the list of excluded patterns (if any) */
  int skip; /* flag set to 1 if pattern is to be skipped */
  int Ytmp, Xtmp;
  FILE *datf;
  char message[MESSAGESIZE];
  char word[WORDSIZE];

  
  if(flag[14]>=2) printf("dataread(): Reading in data\n");

  if(flag[14]>=4) {
    printf("Xtot=%d, Ytot=%d, Xsize=%d, Ysize=%d\n", Xtot, Ytot, Xsize, Ysize);
    for(r=0;r<=xranges[0][0];r++)
      printf("xranges: r=%d  lower=%3d upper=%3d\n", r, xranges[0][r], xranges[1][r]);
    for(r=0;r<=yranges[0][0];r++)
      printf("yranges: r=%d  lower=%3d upper=%3d\n", r, yranges[0][r], yranges[1][r]);
  }

  datf=fopen(datfname,"r");
  if(!datf) {
    sprintf(message, "dataread(): data file %s not found", datfname);
    neterr(message);
  }

  fgets(word, WORDSIZE, datf);
  fgets(word, WORDSIZE, datf);
  fgets(word, WORDSIZE, datf);
  fgets(word, WORDSIZE, datf);
  sscanf(word,"%d %d %d", &Ytmp, &Xtmp, Npats);
  /* check consistency of file headers with specifications */
  if(Ytmp!=Ytot) {
    sprintf(message, "dataread(): Header in data file %s states %d output\n variables in conflict with spec file (which states %d)", datfname, Ytmp, Ytot);
    neterr(message);
  }
  if(Xtmp!=Xtot) {
    sprintf(message, "dataread(): Header in data file %s states %d input\n variables in conflict with spec file (which states %d)", datfname, Xtmp, Xtot);
    neterr(message);
  }

  if(flag[14]>=2) 
    printf("dataread(): Total number of patterns in file = %d\n", *Npats);

  /* allocate memory */

  /* x and tar are deallocated in calling routine, 
     i.e. statnettrain() or statnetapply(). Note that if we are excluding
     vectors, then x and tar will be too large as Npats is only reduced
     to required size after we have read entire data file. This means
     that throughout the program, there will be empty space at the ends
     of x and tar which is not used (it has no effect as Npats is
     equal to the number of used patterns). However, we must deallocate
     the original full size, which will be Npats+flag[18]. */

  *x    = Dmatrix(*Npats,Xsize+1); 
  *tar  = tmatrix(*Npats,Ysize);   

  /* read in data */

  fgets(word, WORDSIZE, datf); /* strip off comment line */
  for(p=0;p<*Npats;++p) {

    fscanf(datf,"%s",word); /* read ID */
    if(flag[18]>0) { /* skip this pattern? */
      skip=0;
      for(e=0;e<flag[18];++e) {
	if(strcmp(word,exclude[e])==0) {
	  skip=1; 
	  if(flag[14]>=2) printf("dataread(): Excluding pattern %s from %s\n", word, datfname);
	  (*Nexcluded)++;
	}
      }
      if(skip==1) { /* read over rest of line, any format acceptable */
	(*Npats)--;
	p--;
	while(fgetc(datf)!=10) /* 10 is ASCII code for newline */
	  ;
	continue;
      }
    }

    for(r=1,k=0,f=1;f<=Ytot;++f) { /* deal with output variables */
      /* f counts over outputs in file; k counts over those we use */
      fscanf(datf,"%s",word);      
      if(f>=yranges[0][r] && f<=yranges[1][r]) { /* add to tar */
	if( !(strcmp(word,"x")) || !(strcmp(word,"X")) ) { 
	  (*tar)[p][k].def = 0;
	  if(flag[14]>=3) {
	    sprintf(message, "dataread(): pattern %d, target variable %d, is undefined in file %s", p+1, f, datfname);
	    netwarn(message);
	  }
	}
	else {
	  (*tar)[p][k].val = atof(word);
	  (*tar)[p][k].def = 1;
	  if(takelog[f]>=-1) { /* take log */
	    if((*tar)[p][k].val<=0.0) {
	      sprintf(message, "dataread(): cannot take log of negative value (pattern %d, output %d", p+1, f);
	      neterr(message);
	    }  
	    (*tar)[p][k].val = log10((*tar)[p][k].val);
	    takelog[f]=k;
	  }
	}
	k++;
      }
      if(f==yranges[1][r]) { /* hit upper range */
	if(r<yranges[0][0])  /* only increment if there are more ranges */
	  r++;
      }
    }
    
    for(r=1,l=0,f=1;f<=Xtot;++f) { /* deal with input variables */
      fscanf(datf,"%s",word);
      if(f>=xranges[0][r] && f<=xranges[1][r]) { /* add to x */
	if( !(strcmp(word,"x")) || !(strcmp(word,"X")) ) {
	  sprintf(message, "dataread(): pattern %d, input variable %d, is undefined in file %s\n statnet cannot cope with missing inputs", p+1, f, datfname);
	  neterr(message);
	}
	else
	  (*x)[p][l] = atof(word);
	l++;
      }
      if(f==xranges[1][r]) { /* hit upper range */
	if(r<xranges[0][0])  /* only increment if there are more ranges */
	  r++;
      }
    }
    
    (*x)[p][Xsize]=INPBIAS;  /* set biases */

  } /* end of this pattern */
  
  fclose(datf);

  if(flag[14]>=2) 
    printf("dataread(): Total number of patterns (after any exclusions) = %d\n", *Npats);
  if(flag[14]>=4) {
    printf("dataread(): Npats=%d\n", *Npats);
    for(p=0;p<*Npats;++p) {
      printf("%2d ", p);
      for(l=0;l<Xsize+1;++l)
	printf("%f ",(*x)[p][l]);
      printf("\n");
    }
  }

}



void writeweights(int Nnets, int Xsize, int Hsize, int Vsize, int Ysize,
		  double ***wtXH, double ***wtHV, double ***wtVY, double **yscale, 
		  double **xscale, double Hlam, double Vlam, char *otwtfname, int *flag)
{
  /* write out weights */
  
  FILE *otwtf; 
  char message[MESSAGESIZE];
  int n,l,m,j,k;

  otwtf=fopen(otwtfname,"w");
  if(!otwtf) { 
    sprintf(message, "writeweights(): weights file %s could not be opened\n",
	    otwtfname);
    neterr(message); /* or could dump weights to stderr as an emergency */
  }

  if(flag[14]>=2)
    printf("writeweights(): Writing weights to %s\n", otwtfname);

  fprintf(otwtf, "# statnet weights file - do not add or remove lines\n");
  fprintf(otwtf, "# #################################################\n");
  fprintf(otwtf, "# N (nets), X (input), H (first hidden), V (second hidden), Y (output): (exc biases)\n");
  if(flag[53]==1)
    fprintf(otwtf, "  %d        %2d          %d                 0                  %d\n", Nnets, Xsize, Hsize, Ysize);
  else
    fprintf(otwtf, "  %d        %2d          %d                 %d                 %d\n", Nnets, Xsize, Hsize, Vsize, Ysize);
  fprintf(otwtf, "# scaling type: \n");
  if(flag[12]==1) {
    fprintf(otwtf, "var\n");
    fprintf(otwtf, "# Y (outputs) mean and stdev scaling factors:\n"); 
    for(k=0;k<Ysize;++k)
      fprintf(otwtf, " % 8.5e % 8.5e\n", yscale[0][k], yscale[1][k]);
    fprintf(otwtf, "# X (inputs) mean and stdev scaling factors:\n"); 
    for(l=0;l<Xsize;++l)
      fprintf(otwtf, " % 8.5e % 8.5e\n", xscale[0][l], xscale[1][l]);
  }
  else if(flag[12]==3)
    fprintf(otwtf, "netsize\n");
  else
    fprintf(otwtf, "none\n");
  fprintf(otwtf, "# Lambda scale parameter for hidden layer:\n");
  fprintf(otwtf, " % 8.6f\n", Hlam);
  if(flag[53]==2) {
    fprintf(otwtf, "# Lambda scale parameter for second hidden layer:\n");
    fprintf(otwtf, " % 8.6f\n", Vlam);
  }

  for(n=0;n<Nnets;++n) {
    fprintf(otwtf, "# ###################### Network %2d ######################\n", n+1);
    if(flag[53]==2)
      fprintf(otwtf, "# wtXH (input-first hidden weights):\n");
    else
      fprintf(otwtf, "# wtXH (input-hidden weights):\n");
    for(l=0;l<Xsize+1;++l) {
      for(m=0;m<Hsize;++m)
	fprintf(otwtf, " % 8.6f ", wtXH[n][l][m]);
      fprintf(otwtf, "\n");
    }
    if(flag[53]==2) {
      fprintf(otwtf, "# wtHV (first hidden-second hidden weights):\n");
      for(m=0;m<Hsize+1;++m) {
	for(j=0;j<Vsize;++j)
	  fprintf(otwtf, " % 8.6f ", wtHV[n][m][j]);
	fprintf(otwtf, "\n");
      }
    }
    if(flag[53]==2)
      fprintf(otwtf, "# wtVY (second hidden-output weights):\n");
    else
      fprintf(otwtf, "# wtHY (hidden-output weights):\n");
    for(j=0;j<Vsize+1;++j) {
      for(k=0;k<Ysize;++k)
	fprintf(otwtf, " % 8.6f ", wtVY[n][j][k]);
      fprintf(otwtf, "\n");
    }
  }  

  fclose(otwtf);
  
}

void evnewname(char *tpnewname, char *datfname)
{
   /* evaluate output file name; cut-off known suffices if present:
      datfname is copied to datfnewname
      if datfnewname ends with INFILESUFFIX then this is removed
      OTFILESUFFIX is appended to datfnewname
      */

  int n;
  char *infsuf  = INFILESUFFIX; /* I presume this allocates memory */
  char *newfsuf = OTFILESUFFIX;
  size_t datfnamelen = strlen(datfname);
  size_t infsuflen  = strlen(infsuf);

  strcpy(tpnewname,"");
  if(strcmp(&datfname[datfnamelen-infsuflen],infsuf)==0) {
    for(n=0;n<(int)(datfnamelen-infsuflen);++n)
      sprintf(&tpnewname[n],"%c",datfname[n]);
  }
  else
    sprintf(tpnewname,"%s.",datfname);
  strncat(tpnewname,newfsuf,strlen(newfsuf));
  
}



/* return weight decay contribution to total error */

double wderr(int Xsize, int Hsize, int Vsize, int Ysize,
	     double **wtXH, double **wtHV, double **wtVY, 
	     double *alpha, int *flag)
{
  double err = 0.0; /* returned by this subroutine */
  int l,k,j,m;

  for(m=0;m<Hsize;++m) { /* not including H bias */
    for(l=0;l<Xsize;++l)
      err += alpha[0]*wtXH[l][m]*wtXH[l][m];
    err += alpha[1]*wtXH[Xsize][m]*wtXH[Xsize][m]; /* input bias */
  }
  for(m=0;m<Hsize+1;++m) { /* including H bias */
    for(j=0;j<Vsize;++j)
      err += alpha[2]*wtHV[m][j]*wtHV[m][j]; 
  }
  for(j=0;j<Vsize+1;++j) { /* including V bias */
    for(k=0;k<Ysize;++k)
      err += alpha[3]*wtVY[j][k]*wtVY[j][k]; 
  }

  err /= 2.0;
  return (err);
}
